/*
 * Goldwater Distributed Web Architecture.
 *
 * Guardian Launcher and Communication.
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.4 $  $Date: 2003/07/19 11:25:24 $
 * 
 * --------------------------------------------------------------------------
 * $Log: bsguardls.c,v $
 * Revision 1.4  2003/07/19 11:25:24  csmith
 * Report to stderr when logfiles cannot be opened
 *
 * Revision 1.3  2003/04/12 18:09:54  csmith
 * SIGCONT didn't work under freeBSD, but SIGHUP did. Hmm.
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.8  2003/01/05 23:00:26  chris
 * Allow running as superuser.  Become desired (effective) user if we're root,
 * or check that we're the desired user if we're not root.
 *
 * Revision 1.7  2002/12/28 16:38:01  chris
 * When returning GWFailure from servers, return gwusrerr not gwerror.
 *
 * Revision 1.6  2002/12/28 00:24:46  chris
 * Through use of lsof discovered files (log and stdout/stderr) that were open
 * multiple times.  Close them before exec'ing servers.
 *
 * Revision 1.5  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.4  2002/08/27 18:56:48  chris
 * Removed pointless logging - if guardian starts, it says so, there is no
 * need for the admin tool to say that it's started too!
 *
 * Revision 1.3  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.2  2002/02/17 23:53:22  chris
 * Use new config code in gwlib.
 *
 * Revision 1.1  2002/02/17 22:31:29  chris
 * New sources - not all of these will be staying...
 *
 * Revision 1.10  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.9  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.8  2001/06/30 15:13:57  chris
 * Get server boot timeout period from registry.
 *
 * Revision 1.7  2001/04/03 18:54:43  chris
 * Cleaned up compile warnings, changed guardian boot log to LOG_SYSINFO.
 *
 * Revision 1.6  2001/03/30 18:46:04  chris
 * Booting now works as single instance, single class or all.
 *
 * Revision 1.5  2001/03/26 10:53:32  chris
 * Changed to use global bbc and _IpcKey variables.
 *
 * Revision 1.4  2001/03/19 16:13:22  chris
 * Introduced proper gw_logf logging.
 *
 * Revision 1.3  2001/03/05 14:34:51  chris
 * Event driven boot/halt of servers.
 *
 * Revision 1.2  2001/01/14 18:40:41  chris
 * Extensive rework.  Remaps STDOUT and STDERR, clears up msg queues, uses
 * sigaction instead of signal.  Passes msgid's of private message queues
 * instead of msgqueue IPC kyes.
 * ---------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/wait.h>

#include <phlib/xl_types.h>

#include <regstrings.h>
#include <bsipc.h>
#include <goldreq.h>
#include <gwlog.h>
#include <gwgbprm.h>
#include <gwreadconf.h>
#include <gwpaths.h>

static int GUARDIAN_SERVER_ID  = 1;
static int GUARDIAN_CLASS_ID   = 1;

static int   wait_for_child   = 0;
static int   guardian_started = 0;

/* ------------------------------------------------------------------------- */
/*
 * Recieves signals from child (guardian) process
 */
static void
child_ready( int sig )
{
  pid_t cpid = -1 ;

  wait_for_child = 0;
  
  switch( sig ) {
  case CHILD_OKAY_SIG: guardian_started = 1;
    break;
    
  case SIGCHLD: 
    cpid = waitpid( -1, NULL, WNOHANG);
    break;
  }
}

/* ------------------------------------------------------------------------- */
/*
 * Launch Guardian Process
 */
RetValue
launch_guardian( int qid, key_t shmipc, char *local_dom, char *dombr )
{
  RetValue ret = Success;
  
  pid_t cpid     =  0;
  char *args[10] = {0};
  char  reg[128] = {0};
  int   i        =  0;
  long  len      =  0;
  int   sout_fd  = 0;
  int   serr_fd  = 0;
  struct sigaction sa, sao;
    
  static char* std_out       = "";
  static char* std_err       = "";
  static char  s_shmipc  [16] = {0};
  static char  s_qid     [16] = {0};
  static char  s_servid  [16] = {0};
  static char  s_classid [16] = {0};

  /* Get (optional) stdout/stderr paths for guardian and servlets
   */
  sprintf( reg, "%s%s%s", R_DOMAIN_B, dombr, R_STDOUT_L );
  if( gw_config_read_char( reg, &std_out, &len ) < 0 ) std_out = "";
  
  sprintf( reg, "%s%s%s", R_DOMAIN_B, dombr, R_STDERR_L );
  if( gw_config_read_char( reg, &std_err, &len ) < 0 ) std_err = "";
  
  /*
   * Convert Servid to a string....
   */
  sprintf( s_shmipc,  "%d", shmipc    );
  sprintf( s_qid,     "%d", qid       );
  sprintf( s_servid,  "%d", GUARDIAN_SERVER_ID );
  sprintf( s_classid, "%d", GUARDIAN_CLASS_ID  );

  wait_for_child = 1;
  guardian_started = 0;

  sa.sa_handler = child_ready;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0; 
  sigaction( SIGCHLD, &sa, &sao );

  sa.sa_handler = child_ready;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0; 
  sigaction( CHILD_OKAY_SIG, &sa, &sao );


  switch( (cpid = fork()) ) {
  case 0:  /* child */
    /* 
     * Initialise Arguments 
     */
    i = 0;
    args[i++] = "guardian";
    args[i++] = local_dom;
    args[i++] = s_servid;
    args[i++] = s_classid;
    args[i++] = s_qid;
    args[i++] = s_shmipc;
    args[i++] = NULL;
    
    /*
     * re-direction of STDOUT and STDERR
     */
	do {
      if( *std_out != '\0' ) {
        if((sout_fd = open( std_out, O_CREAT | O_APPEND | O_RDWR, 0644 )) >= 0){
          close(1);
          dup( sout_fd );
        }
      }
  
      if( *std_err != '\0' ) {
        if((serr_fd = open( std_err, O_CREAT | O_APPEND | O_RDWR, 0644 )) >= 0){
          close(2);
          dup( serr_fd );
        }
      }
    } while(0);

	if( sout_fd < 0 || serr_fd < 0 ) {
      if(sout_fd<0) fprintf( stderr, "Cannot open file %s for writing\n", std_out);
	  if(serr_fd<0) fprintf( stderr, "Cannot open file %s for writing\n", std_err);
	  fprintf( stderr, "Failure to start administrative processes\n" );
	  exit(1);
	}

    /* Close any fd's that we may have open
	 */
    close( sout_fd );
    close( serr_fd );

    gw_closelog();

    /* Move to directory where guardian is to be found.
	 * We do not want to scan the path as we may be running as root.
	 */
    if( chdir(GWSysBin) < 0 ) {
      gw_logf( LOG_SERIOUS, "Cannot change to directory %s", GWSysBin );
	  exit(1);
    } 

    /* Launch - Do not consider path */
    execv( args[0], args );

    kill( getppid(), SIGCHLD ); /* Child failed to start */

    gw_logf( LOG_SERIOUS, "Could not %s", args[0] );

    exit(1); /* We must terminate this child process... */
    break;
    
  case -1: /* Failure */
    gw_logf( LOG_SERIOUS, "Could not exec %s", args[0] );
    ret = Failure;
    break;
    
  default: /* Parent */
    {
      /* Wait for child (the guardian) to either die or signal that
       * it has started (wait for 4 seconds max...)
       */
      int ix = 0;
      while( ix++ < 4 && wait_for_child ) sleep(1);
      if( !guardian_started ) {
	    gw_logf( LOG_SERIOUS, "Failure. Guardian not started" );
	    ret = Success;
      }
    }
    break;
  }

  sa.sa_handler = SIG_DFL;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0; 
  sigaction( SIGCHLD, &sa, &sao );

  sa.sa_handler = SIG_DFL;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0; 
  sigaction( CHILD_OKAY_SIG, &sa, &sao );
  
  return ret;
}

/* ------------------------------------------------------------------------- */
/* end */

