/*
 * Goldwater Bootstrap
 * - Initialise BB
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.12 $  $Date: 2003/09/13 14:07:36 $
 * 
 * --------------------------------------------------------------------------
 * $Log: goldinit.c,v $
 * Revision 1.12  2003/09/13 14:07:36  csmith
 * Helpful start-up messages
 *
 * Revision 1.11  2003/09/13 13:04:32  csmith
 * Portable.net PInvoke helper functions.
 * Guardian bug: Did not clean BBReqQueue table on failure/shutdown of servers.
 * Increased version numbers to reflect significant bug fixes.
 * Announce copyright FSF in administrator.
 * Administrator now pauses 1/10 second between boot/halt commands to give
 * guardian time to breath (not strictly necessary, but may avoid some signal
 * race conditions).
 *
 * Revision 1.10  2003/04/15 10:57:47  csmith
 * Fixed serious bug in Linux sysv ipc resource checking that crept in with the
 * FreeBSD port.
 * Updated example goldwater config files.
 *
 * Revision 1.9  2003/04/13 11:58:37  csmith
 * Do some comparisons on MSGMAX and MSGMNB and choose  MNB if MAX > MNB
 *
 * Revision 1.8  2003/04/12 17:32:27  csmith
 * More FreeBSD fixes
 *
 * Revision 1.7  2003/04/11 10:22:23  csmith
 * FreeBSD port still buggy!
 *
 * Revision 1.5  2003/04/10 21:02:35  csmith
 * Changed parameter passing to fix dereferencing pointer to incomplete type in freebsd
 *
 * Revision 1.4  2003/04/10 20:49:02  csmith
 * Added #include "gw_ipcs.h"
 *
 * Revision 1.3  2003/04/10 20:35:16  csmith
 * ipcs handling for freebsd and linux
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.17  2003/01/07 00:29:08  chris
 * Once we don't need root any more, get rid of it.
 *
 * Revision 1.16  2003/01/05 23:00:26  chris
 * Allow running as superuser.  Become desired (effective) user if we're root,
 * or check that we're the desired user if we're not root.
 *
 * Revision 1.15  2003/01/04 12:36:55  chris
 * made sure umask is <0777
 *
 * Revision 1.14  2003/01/03 00:09:07  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.13  2003/01/02 20:25:30  chris
 * Allows configuration of IPC SHM umask (default is 066)
 *
 * Revision 1.12  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.11  2002/10/20 13:38:30  chris
 * Enabled setting of loglevel from administrator (removed _LogLevel and
 * replaced it with *_LogLevelPtr through out code).
 *
 * Revision 1.10  2002/06/16 11:39:51  chris
 * Changed BB Client list to be the GWDBridge Message Tracking list.
 *
 * Revision 1.9  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.8  2002/05/30 19:35:55  chris
 * Read maxclients from config and configure table in SHM
 *
 * Revision 1.7  2002/05/12 23:00:14  chris
 * When the administrator cannot be started due to Goldwater IPC being
 * unavailable or unstable - log a message at a WARNING or SERIOUS level.
 * Also return the new error GWE_UNSTABLE if the IPC is unstable.
 *
 * Revision 1.6  2002/03/17 18:43:51  chris
 * Handle the destruction of semaphores and shm on exit
 *
 * Revision 1.5  2002/03/17 11:49:07  chris
 * Configuration of spool directory,
 * Reduced lines per page before pager is invoked,
 * Command line shutdown command to do halt of all servers and Guardian.
 *
 * Revision 1.4  2002/03/14 20:45:55  chris
 * Decent server halting!
 *
 * Revision 1.3  2002/03/11 22:34:30  chris
 * More uptodate TODO and other things.
 *
 * Revision 1.2  2002/03/10 15:15:32  chris
 * Admin tool changes
 *
 * Revision 1.1  2002/03/02 15:20:53  chris
 * Startup stuff.
 *
 * Revision 1.4  2002/02/26 00:30:20  chris
 * All new goldwater admin tool source.
 *
 * Revision 1.3  2002/02/24 18:06:37  chris
 * Moved regstrings.h into gwlib.
 * Initialisation puts appdir into BBControl block.
 *
 * Revision 1.2  2002/02/17 23:53:22  chris
 * Use new config code in gwlib.
 *
 * Revision 1.1  2002/02/17 22:31:29  chris
 * New sources - not all of these will be staying...
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>
#include <phlib/xl_linklst.h>

#include "gw_ipcs.h"

#include <mutex.h>
#include <gwgbprm.h>
#include <gwapi.h>
#include <gwlog.h>
#include <gwusers.h>

#include <regstrings.h>
#include <gwreadconf.h>
#include <bbquery.h>
#include <bsguardls.h>
#include <bsipc.h>

#include "goldinit.h"
#include "version.h"
#include "../guardian/version.h" /* Goldwater Version */

const char LOG_NAME[] = "GWAdmin";
const char GWSTARTUP[] = "Goldwater (version "VERSION") Startup\n(c) 2003 Free Software Foundation, Inc.\n\n";

/* Guardians Command Queue
 */
static int   CmdQueueID = -1;

struct msginfo	msginfo_cds;

char *local_dom  = NULL;
char  local_dom_lev[7];

static char  path[128] = {0};
static int   shmid = 0;
static long  config_log_level = 0;

static int      find_domain_branch  ( const char *domain );
static RetValue bs_initialise_log   ( );
static RetValue detect_guardian     ( );
static RetValue initialise_billboard( );
static RetValue bs_invoke_guardian  ( );
static RetValue get_command_queue   ( );
static RetValue init_user           ( );

/* ----------------------------------------------------------------------- */

RetValue
bs_init( AdminMode mode )
{
  RetValue ret = Success;

  long len = 0;
  int  lev = 0;

  /* Find out what the local domain name is
   */
  if( gw_config_read_char( R_LOCAL_DOM, &local_dom, &len ) < 0 ) {
    printf( "Cannot determine local domain name from configuration\n" );
    return Failure;
  }

  /* Find the domain branch node that the local domain may be found...
   */
  lev = find_domain_branch( local_dom );

  if( lev < 0 ) {
    printf( "Cannot find configuration for domain '%s'\n", local_dom );
    return Failure;
  }

  sprintf( local_dom_lev, "%06d", lev ); /* Store string version */

  /* Get and set global _Mask umask value
   */
  if( __gw_umask_from_config( lev ) == Failure ) {
    printf( "Invalid umask in configuration\n" );
    return Failure;
  }

  /* If we're root, become the intended user */
  if( init_user(lev) == Failure ) return Failure;

  if( bs_initialise_log() == Failure ) return Failure;

  /* Depending on the two startup modes, detect the current state of the
   * Goldwater application (whether BB exists and Guardian is running
   * and put things into a consistent state.
   */
  switch( mode ) {
    case GOLDWATER_STARTUP:
      write( 1, GWSTARTUP, sizeof(GWSTARTUP) - 1 );
	  printf( "Initialising local Goldwater administrative processes....\n" );

	  if( (ret = initialise_billboard() ) == Success ) {

	    printf( "Initialising process Guardian....\n" );
	    ret = bs_invoke_guardian();
      }
	break;

    case GOLDWATER_SHUTDOWN:
      ret = detect_guardian(); /* Got to be running to be shut down! */
	break;

	default: /* All other admin commands require Goldwater to be running */
      ret = detect_guardian();
	break;
  }

  /* Now (if mode = startup) that we've started the guardian, we can drop
   * all root privs for ever.  Only bother if we're root. 
   */
  if( !getuid() ) {
    uid_t u;
    gid_t g;

    /* Throw away potential root privs (gotta become root first though) */
    if( ret == Failure || __gw_become_real_user( &u, &g ) < 0 || 
	                      __gw_become_user(u, g) < 0 ) {
	  ret = Failure;
    }
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
bs_remove()
{
  RetValue ret = Success;

  /* Destroy all mutex sets
   */
  destroy_bbc_mutex( );

  ret = destroy_bb( shmid ); /* Remove BB */

  return ret;
}

/* ----------------------------------------------------------------------- */

static RetValue
init_user( int domlev )
{
  uid_t mu;
  uid_t wu;
  gid_t mg;
  gid_t wg;
  char wuser[40];
  char wgroup[40];
  
  /* Who are we? */
  __gw_get_my_ugid( &mu, &mg );
  
  /* Get who we're supposed to be */
  if( __gw_get_user_from_config( domlev, wuser, sizeof(wuser) ) < 0 ) {
    printf( "Username missing in configuration\n" );
    return Failure;
  }
  if( __gw_get_group_from_config( domlev, wgroup, sizeof(wgroup) ) < 0 ) {
    printf( "Usergroup missing in configuration\n" );
    return Failure;
  }

  if( __gw_lookup_ugid( wuser, wgroup, &wu, &wg ) < 0 ) {
    printf( "Unknown User/group given in configuration.\n" );
    return Failure;
  }

  /* If we're root, then become the user we're supposed to be.
   * If we're not root, then barf if we're not who we need to be.
   */
  if( mu ) {
    /* Okay, who are we?
	 */
    if( mu != wu || mg != wg ) {
      gwerror = GWE_BADUSER;
      printf( "User/group running administrator does not match that in configuration.\n" );
      return Failure;
    }
  } else {
    /* Root.  Become wu/wg */
    if( __gw_become_effective_user( wu, wg ) < 0 ) return Failure;
  }

  return Success;
}

/* ----------------------------------------------------------------------- */

static int
find_domain_branch( const char *domain )
{
  int   idx       = 0;
  int   ret       = 0;
  long  len       = 0;
  char *data      = NULL;

  do {
    sprintf( path, "%s%06d%s", R_DOMAIN_B, idx, R_DOM_NAME_L );

    ret = gw_config_read_char( path, &data, &len );

    if( ret >= 0 && strcmp( domain, data ) == 0 ) return idx;

    idx ++;
  } while( ret >= 0 );
  
  return -1;
}

/* ----------------------------------------------------------------------- */
/*
 * Initialise Logging
 */
static RetValue
bs_initialise_log()
{
  long len       = 0;

  sprintf( path, "%s%s%s", R_DOMAIN_B, local_dom_lev, R_LOGFILE_L );

  if( gw_config_read_char( path, &_LogName, &len ) < 0 ) {
    printf( "logfile not set for domain '%s'\n", local_dom );
	gwerror = GWE_BADLOGF;
    return Failure;
  }

  /* Check length of log name
   */
  if( len > FILE_NAME_MAX ) {
    printf( "logfile name too long.  Maximum %d\n", FILE_NAME_MAX );
    return Failure;
  }

  sprintf( path, "%s%s%s", R_DOMAIN_B, local_dom_lev, R_LOGLEVEL_L );

  if( gw_config_read_long( path, &config_log_level ) < 0 ) config_log_level = 0;

  _LogLevelPtr = &config_log_level;
  gw_log_init( 2, LOG_NAME );

  return Success;
}

/* ----------------------------------------------------------------------- */

static RetValue
detect_guardian( )
{
  /* Lookup IPCKey in configuration
   */
  sprintf( path, "%s%s%s", R_DOMAIN_B, local_dom_lev, R_IPCKEY_L );

  if( gw_config_read_long( path, &_IpcKey ) < 0 ) {
    gw_logf( LOG_SERIOUS, "ipckey  not set for domain '%s'\n", local_dom );
	gwerror = GWE_BADIPCKY;
    return Failure;
  }

  /* Create Mutex's (but doesn't force them to be reset.... because if
   * we don't want them cleared every time we start...
   */
  if( create_bbc_mutex( _IpcKey, 0 ) == Failure ) {
    gw_logf( LOG_SERIOUS, "Attach/Creation of BBC Mutex failed " );
	gwerror = GWE_GENFAIL;
    return Failure;
  }

  if( get_bbc_mutex_value( MUTEX_GUARDIAN ) < 1 ) {

    /* Guardian is Running */  

    /* Try and attach to BB. If this fails, we go on to try and recreate it */
    if( (bbc = attach_bb( _IpcKey, &shmid )) == NULL ) {
      gw_logf( LOG_SERIOUS, "Goldwater is unstable. Cannot determine integrity of IPC." );
	  gwerror = GWE_UNSTABLE;
	  return Failure;
    }
  } else {
    gw_logf( LOG_SYSWARN, "Goldwater is not running.  Cannot continue." );
    gwerror = GWE_NOTUP;
    return Failure;
  }

  return Success;
}

/* ----------------------------------------------------------------------- */
/*
 * Read configuration values from the registry, create (or attach to) the 
 * control SHM segment and populate it with config values.
 */
static RetValue
initialise_billboard()
{
  long maxsrv = 0;
  long maxsvc = 0;
  long maxclt = 0;
  long maxdom = 0;
  long len    = 0;
  long os     = 0;
  long msgsize = 0;
  char *appdir = "";
  char *spooldir = "";

  struct msginfo *minfo = NULL;;

  os = sprintf( path, "%s%s", R_DOMAIN_B, local_dom_lev );

  strcpy( path + os, R_IPCKEY_L );
  if( gw_config_read_long( path, &_IpcKey ) < 0 ) {
    gw_logf( LOG_SERIOUS, "ipckey not set for domain '%s'", local_dom );
    gwerror = GWE_BADIPCKY;
    return Failure;
  }

  if( _IpcKey > MAXIPCKEY ) {
    gwerror = GWE_BADIPCKY;
    gw_logf( LOG_SERIOUS, "%s", gw_errorstr(gwerror) );
  }
  
  /* Attach to semaphore set, and if successful, see if Guardian is running.
   * If it is, then we need to fail!
   * If we can't attach to the BB at this point, then we go no further......
   */
  if( attach_bbc_mutex( _IpcKey ) == Success &&
      (bbc = attach_bb( _IpcKey, &shmid )) != NULL ) {

    if( get_bbc_mutex_value( MUTEX_GUARDIAN ) < 1 ) {
      gw_logf( LOG_SERIOUS, "Goldwater Already Running" );
      gwerror = GWE_STILLUP;
      return Failure;
	}

    /* Reset All Mutex Sets */
    if( reset_all_bbc_mutex( ) == Failure ) {
      gw_logf( LOG_SERIOUS, "Initialisation of BBC Mutex set failed" );
      gwerror = GWE_GENFAIL;
      return Failure;
    }

    /* Got mutex set, Guardian is not running....
	 * Now we attach to the BB.  If it's not there ,then we need to go off
	 * and create it.
     */

  } else {
    /* No BB in existance or no mutex set.  Create a new one. */

    gw_logf( LOG_INTDBG, "Attach to BillBoard Failed. Create New." );

    strcpy( path + os, R_MAX_SRVS_L );
    if( gw_config_read_long( path, &maxsrv ) < 0 ) {
      gw_logf( LOG_SERIOUS, "maxservers  not set for domain '%s'\n", local_dom);
      gwerror = GWE_BADCFG;
      return Failure;
    }

    strcpy( path + os, R_MAX_SVCS_L );
    if( gw_config_read_long( path, &maxsvc ) < 0 ) {
      gw_logf( LOG_SERIOUS, "maxservices  not set for domain '%s'\n",local_dom);
      gwerror = GWE_BADCFG;
      return Failure;
    }

    strcpy( path + os, R_MAX_CLIENT_L );
    if( gw_config_read_long( path, &maxclt ) < 0 ) {
      gw_logf( LOG_SERIOUS, "maxclients  not set for domain '%s'\n",local_dom);
      gwerror = GWE_BADCFG;
      return Failure;
    }

    strcpy( path + os, R_EXE_DIR_L );
    if( gw_config_read_char( path, &appdir, &len ) < 0 ) {
      gw_logf( LOG_SERIOUS, "exedir not set for domain '%s'\n", local_dom);
      gwerror = GWE_BADCFG;
      return Failure;
    }

    /* Check length of path
     */
    if( len > FILE_NAME_MAX ) {
      gw_logf( LOG_SERIOUS, "exedir path too long.  Maximum %d\n", FILE_NAME_MAX );
      gwerror = GWE_BADCNFG;
      return Failure;
    }
  
    strcpy( path + os, R_REQ_TOUT_L );
    if( gw_config_read_long( path, &_ReqTout ) < 0 ) _ReqTout = 30;
    
    if( gw_config_read_long( R_MAX_DOMAINS, &maxdom ) < 0 ) {
      gw_logf( LOG_SERIOUS, "maxdomains  not set\n" );
      gwerror = GWE_BADCFG;
      return Failure;
    }

    strcpy( path + os, R_SPOOLDIR_L );
    if( gw_config_read_char( path, &spooldir, &len ) < 0 ) {
	  spooldir = "/tmp/goldspool";
	  len = strlen(spooldir);
    }

    /* Check length of path
     */
    if( len > FILE_NAME_MAX ) {
      gw_logf( LOG_SERIOUS, "spooldir path too long.  Maximum %d\n", FILE_NAME_MAX );
      gwerror = GWE_BADCNFG;
      return Failure;
    }

    if( (bbc = create_bb( _IpcKey, maxsrv, maxsvc, maxclt, maxdom )) == NULL ) {
      gw_logf( LOG_SERIOUS, "Creation of BillBoard Failed" );
      gwerror = GWE_GENFAIL;
      return Failure;
    }

    /* Create New Mutex Set
	 */
    if( create_bbc_mutex( _IpcKey, 1 ) == Failure ) {
      gw_logf( LOG_SERIOUS, "Creation of BBC Mutex failed " );
	  gwerror = GWE_GENFAIL;
      return Failure;
    }
  
    /* Set initial values for bb
	 */
    bbc->loglevel = config_log_level;
    bbc->reqtout  = _ReqTout;
  
    strcpy( bbc->logname,  _LogName );
    strcpy( bbc->appdir,   appdir   );
    strcpy( bbc->spooldir, spooldir );
	bbc->umask = _Mask & 0777;
  }
  
  /* Create a new command Queue for the Guardian, or attach to an existing
   * one. BBC MUST be initialised by this point.
   */
  if( get_command_queue() == Failure ) return Failure;

  /* Stat command queue to find out what the Kernel MSGMAX is configured as.
   * Use this value to calculate the max number of bytes we can send in a
   * single message.
   */
  
  if( __gw_msginfo( CmdQueueID, &minfo ) < 0 ) {
    
    gw_logf( LOG_SERIOUS, "Could not aquire IPC settings" );
    gwerror = GWE_GENFAIL;
    return Failure;
  } 
  else {
	  msgsize = minfo->msgmax;

	/* Verify that the msgqueue ipc settings are sensible....
	 */
    if( msgsize > minfo->msgmnb ) {
      gw_logf( LOG_SYSWARN, "Max bytes per message > max bytes per queue - using max bytes per queue: %ld", minfo->msgmnb );
	  msgsize = minfo->msgmnb;
	}

	if( msgsize <= MsgBlockHdrSize ) {
      gw_logf( LOG_SERIOUS, "Insufficient MSGMAX/MSGMNB (%ld)", msgsize );
      gwerror = GWE_GENFAIL;
      return Failure;
	}

    bbc->maxmsgtx = msgsize - MsgBlockHdrSize;

    gw_logf( LOG_INTDBG, "MSGMAX: %d, MSGMNB: %d, Header: %d, MaxMsgTx: %d",
	     minfo->msgmax, minfo->msgmnb, MsgBlockHdrSize, bbc->maxmsgtx );
  }  

  return Success;
}

/* ----------------------------------------------------------------------- */

static RetValue
bs_invoke_guardian()
{
  gw_logf( LOG_SYSINFO, "Launching Goldwater Guardian" );

  /* Check if Guardian is running or not by checking it's semaphore...
   */
  if( get_bbc_mutex_value( MUTEX_GUARDIAN ) > 0 ) {
    /* Guardian is not running - Launch it
     */

    if( launch_guardian( CmdQueueID, _IpcKey, 
			 local_dom, local_dom_lev  ) == Failure ) {
      gw_logf( LOG_SERIOUS, "Failed to launch Guardian" );
	  gwerror = GWE_GENFAIL;
      return Failure;
    }
  } else gw_logf( LOG_SYSINFO, "Already booted.  Skipping" );
  
  return Success;
}

/* ----------------------------------------------------------------------- */

static RetValue
get_command_queue()
{
  /* Check if Guardian is running or not by checking it's semaphore...
   */
  if( get_bbc_mutex_value( MUTEX_GUARDIAN ) > 0 ) {

    /* Create new queue and attach 
     */
    if( (CmdQueueID = get_private_queue() ) < 0 ) {
      gw_logf( LOG_SERIOUS, "Failed to attatch to IPC" );
      return Failure;
    }
    
    bbc->guardian_msgqid = CmdQueueID;
    
  } else {
    
    CmdQueueID = bbc->guardian_msgqid;
  }


  return Success;
}

/* ----------------------------------------------------------------------- */




/* ----------------------------------------------------------------------- */
/* end */
