/*
 * Goldwater Distributed Web Architecture.
 * - Goldwater Reporting Commands (listing etc)
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:10 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gw_list_cmd.c,v $
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.13  2002/12/01 13:44:30  chris
 * Added Node extension to domain title
 *
 * Revision 1.12  2002/11/30 16:18:29  chris
 * Disable 'pager' for list commands in batch mode.
 *
 * Revision 1.11  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.10  2002/08/26 19:05:09  chris
 * If admin services have not explicitly been asked for in a service list,
 * then do not allow them to be reflected in the total service count.
 *
 * Revision 1.9  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.8  2002/05/21 22:32:18  chris
 * Service list code was breaking out once a class or id was found in
 * the BB greater than that being searched for (if specified with a -c/-i).
 * This prevented services advertied by GWDBridge (from other domains) being
 * listed explicitly.
 *
 * Revision 1.7  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.6  2002/05/12 13:15:45  chris
 * Cleanup of unnecessary fields in BBEntry.
 *
 * Revision 1.5  2002/03/17 11:49:07  chris
 * Configuration of spool directory,
 * Reduced lines per page before pager is invoked,
 * Command line shutdown command to do halt of all servers and Guardian.
 *
 * Revision 1.4  2002/03/16 17:28:49  chris
 * Queue listing.
 *
 * Revision 1.3  2002/03/15 19:36:03  chris
 * Display 'command not implemented' where appropriate.
 * Adjust total number of services to not report those which are private.
 * New version number.
 *
 * Revision 1.2  2002/03/15 19:28:48  chris
 * Added Service List and 'pager' support via more/less ($PAGER)
 *
 * Revision 1.1  2002/03/14 21:49:12  chris
 * Process (server)listing
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>

#include <gwreq.h>
#include <gwlog.h>
#include <gwclient.h>
#include <gwapi.h>

/* Privates */
#include <goldreq.h>
#include <regstrings.h>
#include <gwreadconf.h>
#include <gwgbprm.h>
#include <mutex.h>

#include "gwadmin.h"

#define PAGER         "PAGER"
#define PAGER_DEFAULT "less"

#define INV_CLASS  "ERR: Invalid Class Identifier\n"
#define INV_SERVER "ERR: Invalid Server Identifier\n"
#define SRVL_USAGE "Usage: srvl [pl] -s server|-c class id -i service id|-c class id\n\n"
#define SVCL_USAGE "Usage: svcl [sl] -f function|-s service|-c class id -i service id|-c class id\n\n"

static int dis_pager = 0;
static char force_pager = '\0';

static RetValue list_servers ( long classid, long srvid );
static RetValue list_services( long classid, long srvid, char *f, char*s, int );
static RetValue list_queues  ( long classid, long srvid );
static RetValue list_domains( );

static RetValue pager( CString *textCS );

/* ------------------------------------------------------------------------- */

void
disable_pager( int p )
{
  dis_pager = p;
}

/* ------------------------------------------------------------------------- */

RetValue
gw_process_list( CString *argCS )
{
  long    val    = 0;
  CString optvCS = CSInit;
  char    opt    = '0';
  int     go     = 1;
  long    class  = 0;
  long    id     = 0;
  char    server[128] = {0};

  force_pager = dis_pager ? 'N' : '\0';

  while( CSLenPtr( argCS ) > 0 ) {
    
    if( parse_arguments( argCS, &optvCS, &opt, "s:i:c:aPS" ) == Failure ) {
      /* Give Usage */
      printf( SRVL_USAGE );
      go = 0;
      break;
    }
    
    switch( opt ) {
    case 'c': /* Class */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	printf( INV_CLASS ); /* Error */
	go = 0;
      } else class = val;
      break;
      
    case 'i': /* ID */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	printf( INV_SERVER );
	go = 0;
      } else id = val;
      break;
      
    case 's': /* Server */
      if( CSLen( optvCS ) > sizeof(server) - 1 ) 
	CSLen( optvCS ) = sizeof(server) - 1;
      
      strncpy( server, CSStart(optvCS), CSLen(optvCS) );
      server[ CSLen(optvCS) ] = '\0';
      break;
      
    case 'P': /* Force pager */
      force_pager = 'Y';
      break;

    case 'S':
      force_pager = 'N';
      break;

    case 'a': /* Everything */
      break;
      
    default: 
      printf( SRVL_USAGE );
      go = 0;
      break;
    }
  }
  
  if( id >= 0 && class == -1 && server[0] == '\0' ) {
    printf( INV_CLASS );
    go = 0;
  }
  
  if( go ) list_servers( class, id );
  
  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue
gw_service_list( CString *argCS )
{
  long    val    = 0;
  CString optvCS = CSInit;
  char    opt    = '0';
  int     go     = 1;
  int     all    = 0;
  long    class  = 0;
  long    id     = 0;
  char    service [128] = {0};
  char    function[128] = {0};

  force_pager = dis_pager ? 'N' : '\0';

  while( CSLenPtr( argCS ) > 0 ) {
    
    if( parse_arguments( argCS, &optvCS, &opt, "f:s:i:c:aPSx" ) == Failure ) {
      /* Give Usage */
      printf( SVCL_USAGE );
      go = 0;
      break;
    }
    
    switch( opt ) {
    case 'c': /* Class */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	printf( INV_CLASS ); /* Error */
	go = 0;
      } else class = val;
      break;
      
    case 'i': /* ID */
      if( CStoInt32( &optvCS, &val ) == Failure ) {
	printf( INV_SERVER );
	go = 0;
      } else id = val;
      break;
      
    case 's': /* Service */
      if( CSLen( optvCS ) > sizeof(service) - 1 ) 
	CSLen( optvCS ) = sizeof(service) - 1;
      
      strncpy( service, CSStart(optvCS), CSLen(optvCS) );
      service[ CSLen(optvCS) ] = '\0';
      break;

    case 'f': /* Function */
      if( CSLen( optvCS ) > sizeof(function) - 1 ) 
	CSLen( optvCS ) = sizeof(function) - 1;
      
      strncpy( function, CSStart(optvCS), CSLen(optvCS) );
      function[ CSLen(optvCS) ] = '\0';
      break;
      
    case 'P': /* Force pager */
      force_pager = 'Y';
      break;

    case 'S':
      force_pager = 'N';
      break;

    case 'x': /* ALL services (DEBUG only) */


    case 'a': /* Everything */
	  all = 1;
      break;
      
    default: 
      printf( SVCL_USAGE );
      go = 0;
      break;
    }
  }
  
  if( id >= 0 && class == -1 && service[0] == '\0' ) {
    printf( INV_CLASS );
    go = 0;
  }
  
  if( go ) list_services( class, id, function, service, all );

  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue
gw_queue_list( CString *argCS )
{
  return list_queues( 0, 0 );
}

/* ------------------------------------------------------------------------- */

RetValue
gw_domain_list( CString *argCS )
{
  return list_domains( 0, 0 );
}

/* ------------------------------------------------------------------------- */

static RetValue
list_servers( long classid, long srvid )
{
  RetValue ret   = Success;
  int      l     = 0;
  int      procs = 0;
  uchar   *ptr   = NULL;
  BBEntry *bbe_p = NULL;
  char     buf[1024];
  CString obCS = CSInit;
  CString tCS  = CSInit;

  SHMLIST *BB  = (SHMLIST*)((char*)bbc + bbc->BB_entry_offset );
  BBDom   *bbd = (BBDom*)  ((char *)bbc + bbc->BB_dom_offset);

  /*
   * Get semaphore around shm access
   */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, LLABEL "Could not aquire BBC Mutex" );
    return Failure;
  }
  
  l = sprintf( buf, "Currently booted servers:\n\n"
	       "%-14.14s %-6.6s %-6.6s %-15.15s %-5.5s %-7.7s %-4.4s %-15.15s\n%s\n",
	       "Server", "Class", "ID", "Domain/Node","Pid","Done","FRst","Status",
	       "-------------- ------ ------ "
	       "--------------- ----- ------- ---- ---------------" );
  
  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );
  
  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      bbe_p = (BBEntry*)ptr;
      
      /* Filter on class, ID and server (found in FUNCTION of srch_srv!!!!) */
      
      if( bbe_p->domainid > bbc->domainid ) break;
      
      if( bbe_p->domainid == bbc->domainid ) {

	/* SKIP OVER ADMINISTRATIVE CLASS!!!!!	 */
	/* 	if( bbe_p->bb_sp.servclass == 1 ) continue; */
	
	/* If we want a particular class, and we've overshot, then finish */
	if( classid && bbe_p->bb_sp.servclass > classid ) break;

	/* If we want any class, or a particular class... */
	if( !classid || bbe_p->bb_sp.servclass == classid ) {      

	  /* If we want any server instance, or a particular one... */
	  if( !srvid || srvid == bbe_p->bb_sp.servid ) {

	    l = sprintf( buf,
			 "%-14.14s %-6ld %-6ld %-15.15s "
			 "%-5ld %-7lu %-4d %-15.15s\n",
			 bbe_p->bb_sp.server,
			 bbe_p->bb_sp.servclass,
			 bbe_p->bb_sp.servid,
			 bbd[ bbe_p->domainid - 1 ].domain, /* Cross ref with dom tab (Domain ID's start at 1, array is at ZERO */
			 (long)bbe_p->bb_sc.pid,
			 bbe_p->bb_sc.reqcount,
			 bbe_p->bb_sc.frstrt,
			 bbe_p->bb_sc.state );
	    
	    WrapCStringLen( tCS, buf, l );
	    append_cstring( &obCS, &tCS );
	    
	    procs++;
	  }
	}
      }
    } while( shm_List_get_next( BB, &ptr ) == Success );
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    
    gw_logf( LOG_SERIOUS, LLABEL "Could not release BBC Mutex" );
    ret = Failure;
  }

  l = sprintf( buf, "\n%d servers listed from a total of %d booted.\n\n", procs, shm_List_get_size( BB ) );

  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );

  /* If we've got less than a screen full, then don't bother with pager */
  if( (procs < 18 && force_pager != 'Y' ) || force_pager == 'N' ) {
    write( 1, CSStart(obCS), CSLen(obCS) );
  } else {
    pager( &obCS );
  }

  free_cstring( &obCS );
  return ret;
}

/* ------------------------------------------------------------------------- */

static RetValue
list_services( long classid, long srvid, char *function, char *service, int all)
{
  RetValue ret   = Success;
  int      servc = 0;
  int      l     = 0;
  int      adjust= 0;
  BBSvc   *bbs_p = NULL;
  uchar   *ptr   = NULL;
  char     buf[1024];
  CString obCS = CSInit;
  CString tCS  = CSInit;

  SHMLIST *BBS = (SHMLIST*)((char*)bbc + bbc->BB_svc_offset);
  BBDom   *bbd = (BBDom*)  ((char*)bbc + bbc->BB_dom_offset);

  /*
   * Get semaphore around shm access
   */
  if( aquire_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    gw_logf( LOG_SERIOUS, LLABEL "Could not aquire BBC SVC Mutex" );
    return Failure;
  }

  l = sprintf( buf,
	       "Active Service Table:\n\n"
	       "%-25.25s %-15.15s %-6.6s %-6.6s %-15.15s %-7.7s\n%s\n",
	       "Service", "Function", "Class", "ID", "Domain/Node", "Done",
	       "------------------------- --------------- "
	       "------ ------ --------------- -------" );
  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );


  if( shm_List_get_first( BBS, &ptr ) == Success ) {
    do {
      bbs_p = (BBSvc*)ptr;
      
/* ---
      if( bbs_p->domainid > bbc->domainid ) break;
      if( bbs_p->domainid < bbc->domainid ) continue;
--- */
      
      /* SKIP OVER ADMINISTRATIVE SERVICES!!!!!	 */
      if( !all && bbs_p->service[0] == '!' ) { adjust++; continue; }
      
      if( classid  && bbs_p->servclass > classid              ) continue;
      if( classid  && bbs_p->servclass < classid              ) continue;
      
      if( srvid    && bbs_p->servid    > srvid                ) continue;
      if( srvid    && bbs_p->servid    < srvid                ) continue;
      
      if( *function && strcmp( bbs_p->function, function) != 0 ) continue;
      if( *service  && strcmp( bbs_p->service,  service ) != 0 ) continue;
      
      l = sprintf( buf,
		   "%-25.25s %-15.15s %-6ld %-6ld %-15.15s %-7ld\n",
		   bbs_p->service,
		   bbs_p->function,
		   bbs_p->servclass,
		   bbs_p->servid,
		   bbd[ bbs_p->domainid - 1 ].domain,
		   bbs_p->reqdone );
      
      WrapCStringLen( tCS, buf, l );
      append_cstring( &obCS, &tCS );
      
      servc++;

    } while( shm_List_get_next( BBS, &ptr ) == Success );
  } 
  
  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC_SVC ) == Failure ) {
    
    gw_logf( LOG_SERIOUS, LLABEL "Could not release BBC SVC Mutex" );
    ret = Failure;
  }
  
  /* The adjust stops admin services from affecting the total count
   * if they have not explicitly been asked for
   */
  l = sprintf( buf, "\n%d services listed from a total of %d.\n\n", 
	       servc, shm_List_get_size( BBS ) - adjust );

  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );
  
  /* If we've got less than a screen full, then don't bother with pager */
  if( (servc < 18 && force_pager != 'Y' ) || force_pager == 'N' ) {
    write( 1, CSStart(obCS), CSLen(obCS) );
  } else {
    pager( &obCS );
  }
  
  free_cstring( &obCS );
  return ret;
}

/* ------------------------------------------------------------------------- */

static RetValue
list_queues( long classid, long srvid )
{
  RetValue ret    = Success;
  int      l      = 0;
  int      queues = 0;
  int      qidx   = 0;
  int      o_qidx = -1;
  uchar   *ptr    = NULL;
  BBEntry *bbe_p  = NULL;
  char     buf[1024];
  CString obCS    = CSInit;
  CString tCS     = CSInit;

  BBReqQueue *bbq = (BBReqQueue*)((char*)bbc + bbc->BB_queue_offset);
  SHMLIST    *BB  = (SHMLIST*   )((char*)bbc + bbc->BB_entry_offset);

  /*
   * Get semaphore around shm access
   */
  if( aquire_bbc_mutex( MUTEX_BBC ) == Failure ) {
    gw_logf( LOG_SERIOUS, LLABEL "Could not aquire BBC Mutex" );
    return Failure;
  }

  l = sprintf( buf,
	       "Queue Table:\n\n"
	       "%-6.6s %-6.6s %-8.8s %-8.8s\n%s\n",
	       "Class", "NumOnQ", "AveQTime", "Attached",
	       "------ ------ -------- --------" );

  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );


  if( shm_List_get_first( BB, &ptr ) == Success ) {
    do {
      bbe_p = (BBEntry*)ptr;
      
      /* Filter on class, ID and server (found in FUNCTION of srch_srv!!!!) */
      
      if( bbe_p->domainid < bbc->domainid ) continue;
      if( bbe_p->domainid > bbc->domainid ) break;

      /* SKIP OVER ADMINISTRATIVE CLASS!!!!!	 */
      /* 	if( bbe_p->bb_sp.servclass == 1 ) continue; */
	
      if( classid  && bbe_p->bb_sp.servclass > classid ) break;
      if( classid  && bbe_p->bb_sp.servclass < classid ) continue;
      
      if( srvid    && bbe_p->bb_sp.servid    > srvid   ) break;
      if( srvid    && bbe_p->bb_sp.servid    < srvid   ) continue;

      qidx = bbe_p->req_queue_idx;
      
      if( qidx == o_qidx ) continue; /* Only do each queue once */
      o_qidx = qidx;

      l = sprintf( buf,
		   "%-6ld %-6ld %-8ld %-6ld\n",
		   bbe_p->bb_sp.servclass,
		   bbq[qidx].numonqueue,
		   bbq[qidx].aveQtime,
		   bbq[qidx].numdeps );
      
      WrapCStringLen( tCS, buf, l );
      append_cstring( &obCS, &tCS );

      queues++;

    } while( shm_List_get_next( BB, &ptr ) == Success );
  }

  /* Release Mutex */
  if( release_bbc_mutex( MUTEX_BBC ) == Failure ) {
    
    gw_logf( LOG_SERIOUS, LLABEL "Could not release BBC Mutex" );
    ret = Failure;
  }


  l = sprintf( buf, "\n%d queues listed.\n\n", queues );

  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );

  /* If we've got less than a screen full, then don't bother with pager */
  if( (queues < 18 && force_pager != 'Y' ) || force_pager == 'N' ) {
    write( 1, CSStart(obCS), CSLen(obCS) );
  } else {
    pager( &obCS );
  }

  free_cstring( &obCS );
  return ret;
}

/* ------------------------------------------------------------------------- */

static RetValue
list_domains( )
{
  RetValue ret    = Success;
  int      l      = 0;
  char     buf[1024];
  CString obCS    = CSInit;
  CString tCS     = CSInit;
  int      ix     = 0;
  int      domcount = 0;

  BBDom *domarray = (BBDom*)((char*)bbc + bbc->BB_dom_offset);

  /*
   * Get semaphore around shm access
   */

  l = sprintf( buf, "Domain/Node Table:\n" );

  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );

  domcount = 0;

  for( ix = 0; ix < bbc->dommax; ix++ ) {

    if( *(domarray[ix].domain) != '\0' ) {
      l = sprintf( buf, "\n"
		   "%4ld. ID       : %-s\n"
		   "      Address  : %-s\n"
		   "      Port     : %-5ld\n"
		   "      NameSpace: %s\n"
		   "      Available: %s\n",
		   domarray[ix].domainid,
		   domarray[ix].domain,
		   domarray[ix].gw_ipaddr,
		   domarray[ix].gw_port,
		   domarray[ix].namespace,
		   domarray[ix].available ? "Yes" : "No" );

      
      WrapCStringLen( tCS, buf, l );
      append_cstring( &obCS, &tCS );

      domcount++;
    }
  }

  /* Release Mutex */

  l = sprintf( buf, "\n%d domains listed.\n\n", domcount );

  WrapCStringLen( tCS, buf, l );
  append_cstring( &obCS, &tCS );

  /* If we've got less than a screen full, then don't bother with pager */
  if( (domcount < 5 && force_pager != 'Y' ) || force_pager == 'N' ) {
    write( 1, CSStart(obCS), CSLen(obCS) );
  } else {
    pager( &obCS );
  }

  free_cstring( &obCS );
  return ret;
}

/* ------------------------------------------------------------------------- */

static RetValue
pager( CString *textCS )
{
  int      fd[2];
  pid_t    pid;
  CString  progCS = CSInit;
  CString  tmpCS  = CSInit;
  char    *ep     = NULL;

  if( pipe(fd) < 0 ) {
    gw_logf( LOG_SERIOUS, "Cannot create process pipe" );
    return Failure;
  }

  switch( (pid = vfork()) ) {
  case 0:  /* Child  */
    close( fd[1] );
    if( dup2( fd[0], STDIN_FILENO ) != STDIN_FILENO ) {
      gw_logf( LOG_SERIOUS, "Could not open stdin" );
      exit(1);
    }
    close( fd[0] );

    if( (ep = getenv( PAGER )) == NULL ) {
      ep = PAGER_DEFAULT;
    }
    WrapCString   ( tmpCS, ep );
    chop_line_last( &progCS, &tmpCS, '/' );

    if( execlp( ep, CSStart(progCS), (char*)NULL ) < 0 ) {
      gw_logf( LOG_SERIOUS, "Could not execute '%s'", CSStart(progCS) );
      exit(0);
    }

  case -1: /* Error  */
    gw_logf( LOG_SERIOUS, "Could not open stdin" );
    return Failure;
    
  default: /* Parent */
    close( fd[0] );
    write( fd[1], CSStartPtr(textCS), CSLenPtr(textCS) );
    close( fd[1] );

    if( waitpid(pid, NULL, 0) < 0 ) {
    }
    return Success;
  }

  return Failure;
}

/* ------------------------------------------------------------------------- */
/* end */
