/*
 * Goldwater Bootstrap/Administrator
 *
 * Copyright 1996-2003 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.6 $  $Date: 2003/10/23 21:47:28 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwadmin.c,v $
 * Revision 1.6  2003/10/23 21:47:28  csmith
 * Add goldwater version to goldwater admin welcome message
 *
 * Revision 1.5  2003/09/13 14:07:36  csmith
 * Helpful start-up messages
 *
 * Revision 1.4  2003/09/13 13:04:32  csmith
 * Portable.net PInvoke helper functions.
 * Guardian bug: Did not clean BBReqQueue table on failure/shutdown of servers.
 * Increased version numbers to reflect significant bug fixes.
 * Announce copyright FSF in administrator.
 * Administrator now pauses 1/10 second between boot/halt commands to give
 * guardian time to breath (not strictly necessary, but may avoid some signal
 * race conditions).
 *
 * Revision 1.3  2003/06/14 12:03:48  csmith
 * Allow -v or --version to display the goldwater version info. This allows us
 * to check the version even if goldwater hasn't started (if it had started then
 * we could run the version admin command).
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.25  2003/01/05 23:00:26  chris
 * Allow running as superuser.  Become desired (effective) user if we're root,
 * or check that we're the desired user if we're not root.
 *
 * Revision 1.24  2002/12/01 13:17:12  chris
 * Change precommand execution to use the admin_command function - this will
 * give us the propper prompt> command display and will behave exactly like an
 * admin command.
 *
 * Revision 1.23  2002/12/01 13:10:21  chris
 * Moved where the welcome message is displayed
 *
 * Revision 1.22  2002/12/01 13:08:08  chris
 * Removed force into interactive mode and added options to perform a
 * pre-command before entering batch mode, interactive mode or command mode
 * (yes you can execute two command with goldwater -c 'command1' command2)
 *
 * Revision 1.21  2002/11/30 20:00:01  chris
 * Work to allow forcing into interactive mode.
 *
 * Revision 1.20  2002/11/30 19:40:53  chris
 * Fixed a couple of formatting bugs wrt prompt changing.
 *
 * Revision 1.19  2002/11/30 19:34:38  chris
 * Allow setting of administration prompt.
 *
 * Revision 1.18  2002/11/30 16:18:29  chris
 * Disable 'pager' for list commands in batch mode.
 *
 * Revision 1.17  2002/11/30 15:13:23  chris
 * Added linenumber to error
 *
 * Revision 1.16  2002/11/30 14:59:46  chris
 * Display command being executed in both batch and single command mode.
 *
 * Revision 1.15  2002/11/30 14:52:06  chris
 * Added batch mode for command files on stdin
 *
 * Revision 1.14  2002/11/30 14:49:26  chris
 * Added batch mode from stdin
 *
 * Revision 1.13  2002/11/30 14:20:02  chris
 * Allow single line commands.
 *
 * Revision 1.12  2002/11/30 14:04:42  chris
 * GPL headers
 *
 * Revision 1.11  2002/09/01 11:24:22  chris
 * On shutdown of Goldwater, give Guardian time to die before ripping IPC
 * to shreds.
 *
 * Revision 1.10  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.9  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.8  2002/03/17 18:43:51  chris
 * Handle the destruction of semaphores and shm on exit
 *
 * Revision 1.7  2002/03/17 11:49:07  chris
 * Configuration of spool directory,
 * Reduced lines per page before pager is invoked,
 * Command line shutdown command to do halt of all servers and Guardian.
 *
 * Revision 1.6  2002/03/14 20:45:55  chris
 * Decent server halting!
 *
 * Revision 1.5  2002/03/11 22:34:30  chris
 * More uptodate TODO and other things.
 *
 * Revision 1.4  2002/03/10 15:15:32  chris
 * Admin tool changes
 *
 * Revision 1.3  2002/03/02 15:19:50  chris
 * Removed printfs
 *
 * Revision 1.2  2002/02/28 00:56:13  chris
 * Wasn't initialising BB properly if detection failed for any reason.
 *
 * Revision 1.1  2002/02/26 00:30:20  chris
 * All new goldwater admin tool source.
 *
 * --------------------------------------------------------------------------
 */
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>

#define _GNU_SOURCE
#include <getopt.h>

#include <phlib/xl_types.h>
#include <phlib/xl_string.h>

#include <gwlog.h>
#include <gwreadconf.h>
#include <gwclient.h>
#include <goldreq.h>

#include "command.h"
#include "cmdline.h"
#include "goldinit.h"
#include "cmdline.h"
#include "version.h"
#include "../guardian/version.h" /* Goldwater Version */

int exit_flag = 0;

extern char admin_prompt[];

static const char LOG_NAME[] = "GWAdmin";
const char WELCOME[] = "Welcome to the Goldwater Administrator (version "VERSION"/"ADMIN_VERSION")\n(c) 2003 Free Software Foundation, Inc.\n\nType 'help' or '?' for a list of commands.\n\n";

const char GOODBYE[] = "Goodbye\n";

#define COMMAND_MAX 80
static char precommand[ COMMAND_MAX + 1 ];
CString commandCS = CSInit;

static struct option long_options[] = 
{
  {"silent",      0, 0, 's' },
  {"version",     0, 0, 'v' },
  {"prompt",      1, 0, 'p' },
  {"command",     1, 0, 'c' },
  {0,0,0,0}
};

static int parse_command( CString *inputCS, int lno );
static void interactive();
static void admin_command( char **argv, int max );
static int batch_mode( );
static void shutdown_goldwater();

extern char *local_dom;

/* ----------------------------------------------------------------------- */

int 
main( int argc, char **argv )
{
  int       option_index      = 0;
  char      c                 = '\0';
  char      optstr[]          = "+svp:c:";
  AdminMode mode              = GOLDWATER_ADMIN_INTERACTIVE;
  struct sigaction sa, sao;

  /* Block Ctrl-C and Ctrl-Z
   */
  sa.sa_handler = SIG_IGN;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0;
  sigaction( SIGINT, &sa, &sao );

  sa.sa_handler = SIG_IGN;
  sigemptyset( &sa.sa_mask );
  sa.sa_flags = 0;
  sigaction( SIGTSTP, &sa, &sao );

  strcpy( admin_prompt, "goldwater" ); /* Init prompt */


  /* read arguments here
   */
  while( (c = getopt_long( argc, argv, optstr, 
                           long_options, &option_index )) > (-1) ) {
    switch( c ) {
	  case 'v': /* Display version and exit */
		gw_version(NULL);
		exit(0);
        break;

	  case 's': /* Silent mode */
	    /* silent = 1; */
	    break;

	  case 'p': /* Set prompt string */
        strncpy( admin_prompt, optarg, PROMPT_MAX );
        break;

	  case 'c': /* Execute a command before entering interactive mode */
        strncpy( precommand, optarg, COMMAND_MAX );
		WrapCString( commandCS, precommand );
        break;

	  case ':':
        printf( "Missing Argument\n" );
        exit(1);
        break;

      case '?':
        printf( "Invalid Option\n" );
        break;
	}
  }

  argc -= (optind - 1);

  /* Decide what mode were in */
  if( argc == 1 ) mode = GOLDWATER_ADMIN_INTERACTIVE;
  if( argc >= 2 ) mode = GOLDWATER_ADMIN_COMMAND;

  /* Override mode if we've got data on STDIN */
  if( isatty( STDIN_FILENO ) == 0 )  mode = GOLDWATER_ADMIN_BATCH;

  /* Two special modes which override anything else
   */
  if( argc == 2 ) {
    if( strcmp( argv[optind], "start" ) == 0 ) {
      mode = GOLDWATER_STARTUP;
    } else if( strcmp( argv[optind], "stop" ) == 0 ) {
      mode = GOLDWATER_SHUTDOWN;
	}
  }

  /* Open configuration........ 
   */
  if( gw_config_init() < 0 ) {
    fprintf( stderr, "%s\n", gw_errorstr( gwerror) );
    exit(1);
  }


  /* Initialise system based on mode of operation
   * This includes validating that we're running as the appropriate user, or
   * becoming that user if we're root.
   */
  if( bs_init( mode ) == Failure ) {
    fprintf( stderr, "Error Initialising Administrator (%s)\n", gw_errorstr(gwerror) );
    if( gwerror == GWE_NOTUP ) {
      fprintf( stderr, "[Hint: perhaps you need to run 'goldwater start' first?]\n" );
    }
	exit(1);
  }

  /* Connect to Goldwater ... 
   */
  switch( mode ) {
  case GOLDWATER_SHUTDOWN:
  case GOLDWATER_ADMIN_INTERACTIVE:
  case GOLDWATER_ADMIN_COMMAND:
  case GOLDWATER_ADMIN_BATCH:
    if( gw_init() < 0 ) {
      fprintf( stderr, "Goldwater Client Initialisation Failed (%s)\n", gw_errorstr(gwerror) );
      exit(1);
    }
    gw_log_init( 2, LOG_NAME );
  
    write( 1, WELCOME, sizeof(WELCOME) - 1 );

    break;

    default: break;
  }

  /* If we have a pre-command to execute - do it now
   */
  if( CSLen(commandCS) ) { char *p = precommand; admin_command( &p, 1 ); }

  /* Do tasks based on mode
   */
  switch( mode ) {
  case GOLDWATER_STARTUP: 
	printf( "Goldwater Started.\n" );
    /* Nothing more to do */
    break;

  case GOLDWATER_SHUTDOWN:
    shutdown_goldwater();
    break;

  case GOLDWATER_ADMIN_INTERACTIVE:
    interactive();
    break;

  case GOLDWATER_ADMIN_COMMAND:
    admin_command( argv + optind, argc - 1 );
    break;

  case GOLDWATER_ADMIN_BATCH:
    batch_mode();
    break;
  }

  switch( mode ) {
  case GOLDWATER_STARTUP: 
  case GOLDWATER_SHUTDOWN: 
    break;
  default:
    write( 1, GOODBYE, sizeof(GOODBYE) - 1 );
  }

  gw_config_deinit(); /* Release config */

  exit(0);
}

/* ----------------------------------------------------------------------- */

static void
shutdown_goldwater()
{
  CString     tmpCS = CSInit;
  GWMessage   msg  = {0}; 
  BootParams *rbp  = NULL;
  int         ret  = 0;

  gw_log_init( 2, LOG_NAME );

  /* Halt ALL services */
  gw_halt( &tmpCS );

  /* Halt Guardian */
  printf( "Halting Guardian process....\n\n" );

  printf( "  ++ Halting %-15.15s Class: %-6s ID: %-6s -> ",
	  "guardian", "-", "-" );

  ret = gw_svc_call( "!/gctrl/shutdown", "", 0, &msg, 0 );

  rbp = (BootParams *)msg.data;

  if( ret >= 0 ) {
    printf( "Pid: %-ld ... %.*s\n",rbp->pid,(int)rbp->resp_len,rbp->resp_text);
  } else {
    
    if( gwerror == GWE_APPERR ) {
      printf( "Fail: %.*s\n", (int)rbp->resp_len, rbp->resp_text );
    } else {
      printf( "Fail: %s\n", gw_errorstr( gwerror ) );
    }
  }

  printf( "\n" );

  gw_free_message( &msg );

  /* Give guardian time to die */
  sleep(2);
 
  /* Remove BB and mutexes */
  bs_remove();

  return;
}


/* ----------------------------------------------------------------------- */

static void
interactive()
{
  CString lineCS = CSInit;

  init_readline();

  exit_flag = 0;
  while(!exit_flag) {

    if( get_command_line( &lineCS ) == Success ) {
      if( CSLen(lineCS) ) parse_command( &lineCS, -1 );
    }
  }
}

/* ----------------------------------------------------------------------- */

static void
admin_command( char **argv, int max )
{
  CString lineCS = CSInit;
  int i = 0;

  /* Create a single string for the arguments passed in
   */
  for( i = 0; i < max; i++ ) {
    append_chars( &lineCS, argv[i] );
    append_chars( &lineCS, " "     );
  }

  if( CSLen(lineCS) ) {
    printf( "%s> %.*s\n", admin_prompt, CSLen(lineCS), CSStart(lineCS) );
    parse_command( &lineCS, -1 );
  }
  free_cstring( &lineCS );
}

/* ----------------------------------------------------------------------- */

static int
batch_mode( )
{
  CString lineCS = CSInit;
  CString cmndCS = CSInit;
  char    linebuf[1024];
  int     lineno = 0;
  int     mode   = GOLDWATER_ADMIN_BATCH;
  FILE    *stdin;

  disable_pager( 1 );

  /* Open STDIN */
  stdin = fdopen( STDIN_FILENO, "r" );
  
  while( fgets( linebuf, 1024, stdin ) ) {
    lineno++;
    WrapCString( lineCS, linebuf );
	while( CSLen(lineCS) > 0 && ends_with_chars(&lineCS, "\n")) CSLen(lineCS)--;
	
	/* Skip Comments */
	chop_line( &cmndCS, &lineCS, '#' );
	skip_tail_space( &cmndCS );
	skip_lead_space( &cmndCS );

    if( CSLen(cmndCS) ) {
	  printf( "%s> '%.*s'\n", admin_prompt, CSLen(cmndCS), CSStart(cmndCS) );
	  parse_command( &cmndCS, lineno );
	} 
	linebuf[0] = '\0';
  }
  fclose( stdin );

  return mode;
}

/* ----------------------------------------------------------------------- */
/* Take a command string, detect the command, call the command handler
 * returns:
 *  1: Completed ok
 *  0: Command Failed
 * -1: Command not found
 */
static int
parse_command( CString *inputCS, int lineno )
{
  int ret = -1;
  int i   = 0;
  const char * txt;
  CString cmndCS = CSInit;

  if( chop_word( &cmndCS, inputCS ) == Success ){

    /* Match to command table */
    while( (txt = cmndtab[i].t_text) != NULL ) {

      if( strcmpCS( &cmndCS, txt ) == 0 ) {

	    /* Launch Command Handler */
		ret = 1;
        if( *cmndtab[i].func && (*cmndtab[i].func)( inputCS ) == Failure ) ret = 0;
	    break;
      }
      i++;
    }
	if( ret == -1 ) {
	  printf( "ERR: You have an error in your command near '%.*s'", 
	          CSLen(cmndCS), CSStart(cmndCS) );
      if( lineno >= 0 ) printf( " line %d\n", lineno );
	  else printf( "\n" );
    }
  }

  return ret;
}

/* ----------------------------------------------------------------------- */

RetValue
parse_arguments( CString *inputCS, CString *optval, char *opt, char *opta )
{
  RetValue ret  = Failure;
  int      x    = 0;

  *opt = '\0';

  /* Eat leading space chars
   */
  skip_lead_space( inputCS );
  WrapCStringPtrLen( optval, "", 0 );

  if( CSCharPtr( inputCS, 0 ) != '-' ) return ret; /* Fail */

  *opt = CSCharPtr( inputCS, 1 );

  while( opta[x] != '\0' ) {

    if( opta[x] == ':' ) {
      x++;
      continue;
    }
    
    if( opta[x] == *opt ) {
      string_right( inputCS, CSLenPtr( inputCS ) - 2 );

      if( opta[x+1] == ':' ) {
	/* We're expecting a parameter to follow the option...
	 */
	skip_lead_space( inputCS );
	if( CSCharPtr( inputCS, 0 ) == '-' || 
	    CSLenPtr( inputCS ) < 1 ) return ret; /* Fail */

	/* Get parameter
	 */
	chop_word( optval, inputCS );
      }

      ret = Success;
      break;
    }
    x++;
  }
  
  return ret;
}

/* ----------------------------------------------------------------------- */
/* end */
