#ifndef _GOLD_REQ_H
#define _GOLD_REQ_H
/*
 * Goldwater Request Parameters
 *
 * Copyright 1996-2003 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * ----------------------------------------------------------------------------
 * $Log: goldreq.h,v $
 * Revision 1.4  2003/06/12 21:39:54  csmith
 * Preliminary ASYNC request work.  Pnet C# wrappers.
 *
 * Revision 1.3  2003/04/12 18:09:54  csmith
 * SIGCONT didn't work under freeBSD, but SIGHUP did. Hmm.
 *
 * Revision 1.2  2003/03/29 12:21:10  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.33  2003/03/25 21:16:24  chris
 * Enable auto-recycle of servers
 *
 * Revision 1.32  2003/01/04 16:28:48  chris
 * gw_init is now intelligent about detecting if Goldwater is running.
 * By repeatedly calling gw_init() a client can make sure they stay attached
 * without having to do gw_init(), gw_release() over and over.  gw_init() will
 * detect that the backend has been bounced and try its hardest to reconnect.
 *
 * Revision 1.31  2003/01/03 00:09:11  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.30  2002/08/26 18:57:56  chris
 * Store domain array idx in shm instead of the real msgqueue id.
 *
 * Revision 1.29  2002/08/01 21:06:38  chris
 * Updated BBDMsgTrack with time tracking and removed class/id
 *
 * Revision 1.28  2002/07/28 18:27:12  chris
 * Further integration with netserver.
 *
 * Revision 1.27  2002/06/16 11:40:35  chris
 * BB Client is now the Domain Bridge message tracking list.
 * New error codes and messages.
 *
 * Revision 1.26  2002/06/15 18:54:39  chris
 * Domain Bridge additions
 *
 * Revision 1.25  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.24  2002/05/30 19:37:34  chris
 * Create entries in BBClient table for each new client attaching to
 * application.
 *
 * Revision 1.23  2002/05/26 15:32:50  chris
 * Removed 'local' indicator from BBSvc and replaced it with 'flags'
 *
 * Revision 1.22  2002/05/21 21:15:16  chris
 * Big tidy up - lots of misc changes.
 * Changed 'internal' services to be prefixed by '!' so they don't show in
 * the admin tool service list.
 *
 * Revision 1.21  2002/05/12 17:46:16  chris
 * Store msgid of expected reply in SHM so that Domain Bridge can validate
 * the reply at it's door ..
 *
 * Revision 1.20  2002/05/12 16:33:40  chris
 * Added RequestCtl
 *
 * Revision 1.19  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.18  2002/05/12 13:16:22  chris
 * Cleanup of unnecessary fields in BBEntry.
 * (Created new structure ServerParams that contains the fields common
 *  to those stored in the BB and those in the BootParam message).
 *
 * Revision 1.17  2002/05/11 13:40:44  chris
 * Clean up - for Domain handling.
 *
 * Revision 1.16  2002/03/17 11:19:21  chris
 * Different lens for service name and function
 *
 * Revision 1.15  2002/03/12 22:51:03  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.14  2002/03/10 19:35:33  chris
 * Handles proper homed messages (initially for shutdown of servers).
 *
 * Revision 1.13  2002/03/02 15:22:08  chris
 * Added new fields to BootParams
 *
 * Revision 1.12  2002/02/24 18:05:38  chris
 * Moved APPDIR to BBControl block instead of duplicating it for each and
 * every server.
 * Return GWE_APPERR and set gwusrerr when Services return failure.
 * Use new regstrings that describe the new XML based configuration.
 * Added new functions to support individual addition of service entries.
 * New Version 1.14
 *
 * Revision 1.11  2002/01/03 20:27:31  chris
 * Introduction of preliminary Domain Control handling.
 *
 * Revision 1.10  2001/12/26 22:57:42  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.9  2001/04/15 12:57:09  chris
 * Introduced domain table.
 *
 * Revision 1.8  2001/04/01 22:44:18  chris
 * Removed obsolete guardian control commands and added guardian shutdown
 * command.
 *
 * Revision 1.7  2001/03/31 18:33:58  chris
 * Added repmqid to ServerCtrl.
 *
 * Revision 1.6  2001/03/26 18:45:13  chris
 * Removed CMD_SRV_SHUT_C, added CMD_SRV_SHUT_R, removed entry_num from
 * ServerCtrl and added pid.  Added listen_ident to BBEntry.
 *
 * Revision 1.5  2001/03/24 10:37:35  chris
 * Added guardian_msgqid.
 *
 * Revision 1.4  2001/03/19 15:43:33  chris
 * Lots of new fields.  reqdone and picked are now unsigned longs.
 * gw_ipaddr is a string not a long.
 *
 * Revision 1.3  2001/03/05 14:22:31  chris
 * added service map list.
 * Additional service structure parameters.
 *
 * Revision 1.2  2001/01/14 18:47:51  chris
 * Changed reqipc to msgqid in BBSvc
 *
 * Revision 1.1  2001/01/14 18:25:42  chris
 * Initial revision
 *
 * ----------------------------------------------------------------------------
 */
#include <sys/types.h>

#include <phlib/xl_memlst.h>
#include <gwreq.h>

#define DOM_NM_MAX     31     /* Domain Name Max  */
#define SRV_NM_MAX     63     /* Server Name Max  */
#define HOST_NM_MAX    63     /* Host Name Max    */
#define NAME_SPACE_MAX 63     /* Namespace Max    */
#define FILE_NAME_MAX  127    /* File Name Max    */
#define TXID_MAX       32     /* Max width of txid */

/* For some reaonn FreeBSD does not respond to the SIGCONT signal..
 * Only tried on FreeBSD5.0 though... Most Odd.
 * If anyone finds out why, please let me know! 
 */
#if defined(__FreeBSD__)
#  define CHILD_OKAY_SIG SIGHUP
#else
#  define CHILD_OKAY_SIG SIGCONT
#endif

typedef struct {
  int  numbooted;      /* Number of servers booted */
  int  servmax;    
  int  svcmax;
  int  dommax;
  int  accessmax;
  int  queuemax;

  uuid_t instance_id;    /* The ID of the booted instance           */

  long guardian_msgqid;  /* Guardians Request Message Queue         */
  long dbridge_msgq_idx; /* Domain Bridge Request Message Queue IDX */
  long domainid;         /* Index of Domain entry for Local Dom     */
  mode_t umask;          /* IPC umask */

  char appdir  [FILE_NAME_MAX + 1];
  char spooldir[FILE_NAME_MAX + 1];
  char logname [FILE_NAME_MAX + 1];
  long loglevel;

  long phlrglimit;     /* Large Limit for NetServer */
  long maxmsgtx;       /* Large Limit for SHM Msgs  */

  long reqtout;        /* Request Timeout           */

  char gwhost  [HOST_NM_MAX + 1]; /* IP   of NetServer */
  long gwport;                    /* Port of NetServer */

  int  BBSize;                   
  int  BB_entry_offset; /* Offset to BBEntry     List  */
  int  BB_svc_offset;   /* Offset to BBSvc       List  */
  int  BB_dtrack_offset;/* Offset to BBDMsgTrack List  */
  int  BB_dom_offset;   /* Offset to BBDom       Array */
  int  BB_queue_offset; /* Offset to BBQueue     Array */

} BBControl; /* Server Status Control */

typedef struct {
  long  servid;        /* Server Id            */
  long  servclass;     /* Server Class Id      */
  long  max;           /* Max Servers to boot  */
  long  min;           /* Min Servers to boot  */
  long  rstrt_thresh;  /* Restart Threshold    */
  long  f_rstrt_limit; /* Forced Restart Limit */
  char  server[SRV_NM_MAX + 1]; /* Server name */
} ServerParams;

typedef struct {
  ServerParams sp;
  long  pid;                   
  char  domain[DOM_NM_MAX + 1];
  long  resp_len;
  char  resp_text[1]; /* This MUST BE FOLLOWED by extra memory if resp_len>0 */
} BootParams;


/*
 * One ServerCtrl per server booted
 */
typedef struct {
  pid_t         pid;      /* Server PID                             */
  unsigned long reqcount; /* Requests satisfied by server           */
  int           frstrt;   /* How many forced restarts of server     */
  int           respawn;  /* Whether server gets respawned on death */
  long          gracerstrt_ind; /* Graceful Restart Ind */
  char          state[ SVC_NM_MAX + 1]; /* Current State of server  */
  int           repmqid;  /* Reply msg queue id, if created         */
  long          expect_msgid; /* ID expected of next reply          */
} ServerCtrl;


/* One BBEntry per server booted, and each instance of..
 *
 * - Contains SERVER specific information.  Dynamic and static.
 */
typedef struct {
  long         req_queue_idx; /* Idx of request Queue entry for server  */
  long         domainid;
  ServerCtrl   bb_sc;
  ServerParams bb_sp;
} BBEntry;


/* Queue Handling
 */
typedef struct {
  long msgqid;        /* ID of message Queue (request)             */
  long listen_ident;  /* Base Listen ID for messages on this queue */
  long numonqueue;    /* Number of messages pending in queue       */
  long aveQtime;      /* Average time on queue (ms)                */
  long numdeps;       /* Number of Servers dependent on this queue */
} BBReqQueue;


/* Billboard Entry
 */

/*
 * One BBSvc for each service and alias of
 *
 * - Contains SERVICE specific information.  Dynamic and Static.
 */
typedef struct {
  char  service [ SVC_NM_MAX + 1 ];
  char  function[ FNC_NM_MAX + 1 ];
  long  domainid;
  long  servclass;
  long  servid;
  long  req_queue_idx;   /* Idx of request Queue entry of owning server      */
  ServiceFlags flags;    /* Flags .... like local/remote etc                 */
  unsigned long reqdone; /* Number of requests done by this service (success)*/
  unsigned long picked;  /* Number of times this request has been picked     */

} BBSvc;

/*
 * One for each domain currently know about..
 */
typedef struct {
  long  domainid;
  char  domain   [ DOM_NM_MAX + 1 ];
  char  namespace[ NAME_SPACE_MAX + 1];
  int   available; /* Whether available or not (avail may be transient)      */
  char  gw_ipaddr[ HOST_NM_MAX + 1 ];
  long  gw_port;   /* Sched port */
  
  /*
   * LDBal Loading stats...
   */
} BBDom;

/*
 * Domain Message Tracking - Used between DomainBridge and netServer
 */
typedef struct {
  uuid_t expect_msgid; /* ID of expected reply message */
  uuid_t txid;         /* If we time out, then we need to expire the tx too */
  int    repmqid;      /* Reply msg queue id, if created      */
  time_t time_sent;    /* Local time message was sent         */
  time_t timeout;      /* Local time that the message expires */
} BBDMsgTrack;

#endif
