/*
 * Goldwater Distributed Web Architecture.
 *
 * Message Processing API
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 *
 * $Revision: 1.7 $  $Date: 2003/06/12 21:39:54 $
 * 
 * --------------------------------------------------------------------------
 * $Log: gwmsg.c,v $
 * Revision 1.7  2003/06/12 21:39:54  csmith
 * Preliminary ASYNC request work.  Pnet C# wrappers.
 *
 * Revision 1.6  2003/05/29 22:39:17  csmith
 * Added some TODO's for async handling
 *
 * Revision 1.5  2003/05/10 10:38:34  csmith
 * Fixed conflicts
 *
 * Revision 1.4  2003/05/06 09:31:31  csmith
 * Spooled reply messages were not being colelcted properly as the total size
 * of the payload was not being set from the ctrl message.
 *
 * Revision 1.3  2003/04/13 12:00:54  csmith
 * More debugging
 *
 * Revision 1.2  2003/03/29 12:21:11  csmith
 * cvs merge to 0.3.4
 *
 * Revision 1.35  2003/01/03 00:09:11  chris
 * Use internal umask when creating files, shm and queues
 *
 * Revision 1.34  2002/11/10 12:23:41  chris
 * Cleanup and low level API changes for bindings such as C#
 *
 * Revision 1.33  2002/08/27 18:51:57  chris
 * Spooled replies have been broken since uuids were used as msgids.
 * uuid_matches return 0 if there are no differences and NOT TRUE!
 *
 * Revision 1.32  2002/08/26 00:29:54  chris
 * More control for domain handling (like forcing the removal of spoolfiles)
 *
 * Revision 1.31  2002/08/05 00:42:17  chris
 * Changed __gw_set_requestCtrl to take new and pass back copy of old - this
 * allows buffers to be swapped around the internal req_ctrl.
 * ie given parameters &A and &B, Perform: B <- req_ctrl <- A
 *
 * Revision 1.30  2002/08/04 19:36:03  chris
 * query_bb now allows scanning of a single domain's services.
 * API changes to allow replies to be sent back from dombridge.
 *
 * Revision 1.29  2002/07/28 18:27:12  chris
 * Further integration with netserver.
 *
 * Revision 1.28  2002/06/18 00:00:29  chris
 * destroy_bbc_mutex was coded as though individual semaphores in the set were
 * being destroyed, not the set as a whole.  This caused IPC_RMID errors.
 * Tidied loggin in gwmsg.c.
 *
 * Revision 1.27  2002/06/16 00:19:43  chris
 * Removed debugging from gwuuid_compare
 *
 * Revision 1.26  2002/06/16 00:18:31  chris
 * Use gwuuid_compare to compare uuid_t buffers (4 32bit ints) because it is
 * faster.
 *
 * Revision 1.25  2002/06/15 19:32:05  chris
 * Message ID values are now 128 bit OSF DCE compatible uuid's.
 *
 * Revision 1.24  2002/06/11 21:52:58  chris
 * Enabled timed blocking in gwp_get_request and use this in dbridge to
 * prevent spinning and chewing cpu time.
 *
 * Revision 1.23  2002/06/10 17:34:47  chris
 * All phlib includes have phlib/ prefix and gwlib includes have nothing.
 *
 * Revision 1.22  2002/06/09 20:19:26  chris
 * More TODO and comment changes
 *
 * Revision 1.21  2002/06/04 16:35:45  chris
 * Introduce target_dom
 *
 * Revision 1.20  2002/05/13 17:57:35  chris
 * Beginnings of NEW domain support.
 *
 * Revision 1.19  2002/05/12 14:03:13  chris
 * GPL licence headers etc
 *
 * Revision 1.18  2002/05/11 13:40:44  chris
 * Clean up - for Domain handling.
 *
 * Revision 1.17  2002/04/21 20:20:35  chris
 * Logging additions
 *
 * Revision 1.16  2002/03/17 11:38:20  chris
 * Make sure we've got a pointer to a valid data area (ie not NULL) even
 * if the request message length is ZERO.
 *
 * Revision 1.15  2002/03/17 11:17:45  chris
 * Different sizes for function and service names (serivce => func)
 *
 * Revision 1.14  2002/03/12 22:51:03  chris
 * Added BBReqQueue support.  Currently defines a BBReqQueue entry for
 * each server, even though servers are grouped into classes and thus
 * will usually share a queue.
 *
 * Revision 1.13  2002/03/11 22:25:48  chris
 * Bugfixes: Getting in a twist when collecting stale reply packets and
 * corrupting the ctrl structure.  Fixed stale packet loop logic and made
 * sure ctrl struct side effects are documented and considered.
 *
 * Revision 1.12  2002/03/10 19:35:33  chris
 * Handles proper homed messages (initially for shutdown of servers).
 *
 * Revision 1.11  2002/03/04 00:23:58  chris
 * More internal debugging
 *
 * Revision 1.10  2002/03/02 19:09:39  chris
 * Handle skipping of stale messages in the queue.
 *
 * Revision 1.9  2002/01/29 22:43:40  chris
 * Cleaned up warnings.
 * Added gwp_ prefix to shutdown_server.
 * Added regdata.[ch] config entry to build.
 *
 * Revision 1.8  2001/12/26 22:57:43  chris
 * Changed all references to Imagepage Computing to netFluid Technology.
 *
 * Revision 1.7  2001/04/10 11:47:52  chris
 * As we only ever receive one msgqueue block per request/reply, we do not
 * need to create a new cstring if the data request is too big to fit into
 * our existing buffer.  We just reset the cstring.
 *
 * Revision 1.6  2001/04/01 22:43:25  chris
 * Use different LrgMsgThresholds for Phoenix and msgqueue bound data.
 * Store reply queue ID in SHM to facilitate cleanup on server
 * termination.  Moderate changes to when ctrl->bsize is set (now in
 * write_block_init).
 *
 * Revision 1.5  2001/03/26 18:40:35  chris
 * Handles proper shutdown of servers via the servers message queue.  The
 * server sends a Guardian Style Command back to the originator of the
 * shutdown message (ie guardian), pauses for a 'random' number of
 * milliseconds and then exits.
 *
 * Revision 1.4  2001/03/26 12:51:35  chris
 * Removed continuation message blocks and introduced dest_srv_id header
 * parameter.
 *
 * Revision 1.3  2001/03/26 10:44:34  chris
 * gwp_get_request now transfers chain_ident from mb to ctrl.
 *
 * Revision 1.2  2001/03/22 17:05:00  chris
 * Comment change.
 *
 * Revision 1.1  2001/03/19 16:05:56  chris
 * Initial revision
 *
 * --------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <errno.h>

#define GW_INTERNAL_FLAGS 1

#include <gwgbprm.h>
#include <goldreq.h>
#include <gwtime.h>
#include <gwmsg.h>
#include <gwlog.h>
#include <gwportal.h>

const int ReqBlockTime = 5;

/* ------------------------------------------------------------------------- */

static int
gwp_send_message( GWMsgCtrl *ctrl, const char *data, long len )
{
  long     msize  = 0;
  int      flags  = 0;
  MsgBlock mb     = {0};

  /*
   * First message to be sent must have the correct MTYPE value
   * This first message contains the continuation mtype to be used by the
   * recipient to get the remaining messages in the chain.
   */
  
  gw_logf( LOG_INTDBG, "gwmsg.c: Send message - (packet: %ld, msg: %ld)", 
           len, ctrl->total_size );
  
  mb.ident          = ctrl->ident;
  mb.total_size     = ctrl->total_size;
  mb.rep_msgqid     = ctrl->r_msgqid;
  mb.org_rep_msgqid = ctrl->org_r_msgqid; /* Used by replies from domain call */
  mb.dest_srv_id    = ctrl->destsrvid;  
  mb.dest_dom_id    = ctrl->target_dom;  
  mb.msgtype        = ctrl->msgtype;
  
  mb.clid           = ctrl->clid;
  uuid_copy( mb.msgid, ctrl->msgid );
  
  strncpy( mb.service, ctrl->service ? ctrl->service : "", SVC_NM_MAX );
  mb.service[ SVC_NM_MAX ] = '\0';
  
  strncpy( mb.function, ctrl->function ? ctrl->function : "", FNC_NM_MAX );
  mb.function[ FNC_NM_MAX ] = '\0';
  
  mb.dsize       = len;
  
  mb.flags       = ctrl->flags;
  
  memcpy( &mb.data, data, len );
  
  msize = ((long)&mb.data - (long)&mb.flags) + len;
  
  gw_logf( LOG_INTDBG, "   * Write Message With IDENT %ld, Q %ld, S %ld, F %x *", mb.ident, ctrl->msgqid, msize, flags );

  if( msgsnd( ctrl->msgqid, (struct msgbuf*)&mb, msize, flags ) < 0 ) {
    gw_logf( LOG_SERIOUS, "Message send failure. (Errno %ld)", errno );
    gwerror = GWE_SYSFAIL;
    return -1;
  }

  return 0;
}

/* ------------------------------------------------------------------------- */

RetValue
gwp_write_block_init( GWMsgCtrl *ctrl  )
{
  int   fd    = 0;
  long  large = 0;
  int   is_a_portal_request = 0;

  ctrl->cflags        = 0;
  ctrl->bytes_spooled = 0;
  ctrl->spfileptr     = NULL;

  /* If we're sending the msg via Phoenix (req or reply) then set large
   * threshold to be NetServer block size.
   * Else set it to be _MaxMsgTx.
   */
  if( (ctrl->flags & GWP_REMOTE) ||
      ((ctrl->flags & GWP_REPLY) && (ctrl->flags & GWP_PHXRET) ) ) {

    large = 4096; /* Get From Reg */

  } else {
    large = _MaxMsgTx;
  }
  
  /*
   * If data exceeds our large file threshold, then mark it as such
   */
  gw_logf( LOG_INTDBG, "Large file threshold is %ld", large );

  if( ctrl->total_size > large ) ctrl->flags |= GWP_LRGMSG;
  
  /* If we're a local request and we want replies to come back via msg queue
   * and not via NetServer, then we'll need a reply queue....
   * [But not if we've said NOREPLY of course!]
   */
  /* TODO - ASYNC - If the request is ASYNC, and we're calling a remote
   * service, where is the appropriate place to set the reply queue?
   * We want reples to go back the that AsyncMgr to be queued.
   */
  if( !(ctrl->flags & GWP_REMOTE) && !(ctrl->flags & GWP_REPLY) 
      && !(ctrl->flags & GW_NOREPLY) && !(ctrl->flags & GWP_PHXRET ) ) {
    
    /* If the message came in from the Domain NetServer, then _ReplyQueue
     * will have been set to the REQUEST QUEUE of the DomainBridge - this is
     * the 'exit' point of the domain!
     */
    if( _ReplyQueue < 0 ) {
      gw_logf( LOG_INTDBG, "Creating Reply Queue" );
      _ReplyQueue = msgget( IPC_PRIVATE, IPC_CREAT | (0666 & ~_Mask) );
      if( _ReplyQueue < 0 ) {
	gwerror = GWE_SYSFAIL;
	return Failure;
      }

      /* ----====++++ THIS ONLY APPLIES TO SERVERS ++++====---- */
      
      if( _MyServEntry != NULL ) {
	/* If we're a server, then store our reply queue id in the SHM
	 * to facilitate cleanup in case things go wrong later
	 */
	_MyServEntry->bb_sc.repmqid = _ReplyQueue;
      }
      
      /* ----====++++OOOOOOOOOOOOOOOOOOOOOOOOOOOOOO++++====---- */

    }

    /* TODO If this is an ASYNC request, then replies will go to the
     * Async Manager and NOT to the requesting client
     */
    ctrl->r_msgqid = _ReplyQueue;

  }

  if( (ctrl->flags & GWP_REMOTE) ||
      ((ctrl->flags & GWP_REPLY) && (ctrl->flags & GWP_PHXRET)  ) ) {
    /* if (remote) OR (reply AND Phoenix) then send to Phoenix via gwportal
     */
    ctrl->cflags |= GWC_PORTAL;
    gw_logf( LOG_INTDBG, ">> Message to go via portal [%s:%d] <<",
	     ctrl->host, ctrl->port );
    
    if( gwp_portal_connect( ctrl->host, ctrl->port ) < 0 ) {
      return Failure;
    }
    
    ctrl->flags &= ~GWP_REMOTE;
    
    if( !(ctrl->flags & GWP_REPLY) ) is_a_portal_request = 1;
  }
  
  /*
   * Spool if we're an ASYNC REPLY   -or- 
   *                   LARGE MESSAGE and NOT A NETSERVER REQUEST
   */
  if( ( (ctrl->flags & GWP_REPLY)  &&  (ctrl->flags & GWP_ASYNC) ) || 
      ( (ctrl->flags & GWP_LRGMSG) && !is_a_portal_request ) ) {

    
    /* if Async reply OR large message, then we need to spool */
    
    ctrl->cflags |= GWC_SPOOL;
    ctrl->flags  |= GWP_DFILE;
 
    gw_logf( LOG_INTDBG, ">> Message to be spooled <<" );

    sprintf( ctrl->spoolfile, "%s/%sXXXXXX",
	     ctrl->spool_dir, SPOOL_FILE_PREFIX );

    /* Use mkstemp to create and open a temporary file....
     */
    if( (fd = mkstemp( ctrl->spoolfile )) < 0 ) {
      if( errno == ENOENT ) {
        gw_logf( LOG_INTDBG, "Spool directory does not exist.  Creating." );
	
	if( mkdir( ctrl->spool_dir, 0775 ) == 0 ) {
	  /* Directory created ok.... Have another go at writing file.. */
          gw_logf( LOG_INTDBG, "Attempting write of spool file again." );
	  
          sprintf( ctrl->spoolfile, "%s/%sXXXXXX",
		   ctrl->spool_dir, SPOOL_FILE_PREFIX );
          fd = mkstemp( ctrl->spoolfile );
	} 
      }
    }
    if( fd < 0 ) {
      gw_logf( LOG_SERIOUS, "Cannot create spoolfile '%s' (errno: %ld)",
	       ctrl->spoolfile, (long)errno );
      gwerror = GWE_SPLDIR;
      return Failure;
    }
    
    /* Change permissions on file as mkstemp probably created it as 0600
     */
    fchmod( fd, 0666 & ~_Mask );
    
    gw_logf( LOG_INTDBG, "Created spoolfile '%s'", ctrl->spoolfile );
    
    /* Make space in the file
     */
    lseek( fd, ctrl->total_size - 1, SEEK_SET );
    write( fd, "", 1 );

    /* Memmap the file
     */
    ctrl->spfileptr = (unsigned char *)mmap( 0, ctrl->total_size,
					     PROT_READ | PROT_WRITE,
					     MAP_SHARED, fd, 0);
    close(fd);
    
    if( ctrl->spfileptr  < 0 ) return Failure;
  } else {
    ctrl->flags &= ~GWP_DFILE;
  }

  ctrl->bsize = _MaxMsgTx; /* Initialise block size */

  reset_cstring( &ctrl->cache );

  return Success;
}

/* ------------------------------------------------------------------------- */
/*
 * Performs buffering of data until bsize bytes are available, they are then
 * sent.  If writing to gwportal, then entire data buffer is sent at once.
 */
RetValue
gwp_write_block_data( GWMsgCtrl *ctrl, const char *data, long dsize )
{
  CString tmpCS = CSInit;

  /* If we're done, then just return...
   */
  if( ctrl->cflags & GWC_DONE ) return Success;

  WrapCStringLen( tmpCS, data, dsize);

  if( ctrl->cflags & GWC_PORTAL && !(ctrl->cflags & GWC_SPOOL) ) {
    /* We need to send data IMMEDIATELY to gwportal, without spooling.
     * Caching may still be necessary...
     */

    if( CSLen(ctrl->cache) + dsize < ctrl->total_size ) {
      /* We need to add data to our cache */
      append_cstring( &(ctrl->cache), &tmpCS );
      gw_logf( LOG_INTDBG, "A: Appending data to cache" );

    } else {
      /* We've got all the data to send, let's do it! */

      ctrl->flags &= ~GWP_REMOTE;  /* Clear REMOTE flag */
      if( CSLen(ctrl->cache) > 0 ) {
	/* Add new data block to cached data, then send cache */
	append_cstring( &(ctrl->cache), &tmpCS );
	gw_logf( LOG_INTDBG,
		 "B: Appending data to that in cache and sending to Sched" );

	if( gwp_send_to_portal( ctrl, CSStart(ctrl->cache), 
				CSLen(ctrl->cache) ) < 0 ) {
	  gw_logf( LOG_SERIOUS, "Send_to_portal failed with code %d", gwerror);
	  return Failure;
	}

	ctrl->cflags |= GWC_DONE;

      } else {
	/* data block is the whole data stream.  We've not had to cache
	 * anything!  Woo Hoo!  Send this block!
	 */
	gw_logf( LOG_INTDBG, "Sending data to scheduler" );

	if( gwp_send_to_portal( ctrl, data, dsize ) < 0 ) {
	  gw_logf( LOG_SYSERR, "Send_to_portal failed with code %d", gwerror);
	  return Failure;
	}
	ctrl->cflags |= GWC_DONE;
      }
    }
  } else {
    /* process data blocks, adding to cache and processing cache in bsize
     * chunks.
     *
     * If we're spooling data to disk, then write data straight to disk
     * cache.  When total_size bytes written (ie all of the data) we can
     * send message to recipient (via queue or gwportal)
     *
     * If we're NOT spooling to disk, append data to cache and process
     * cache in bsize chunks, sending each as a message to recipient.
     */
    if( ctrl->cflags & GWC_SPOOL ) {
      /*
       * write data to disk_file.
       */
      gw_logf( LOG_INTDBG, "Spool to disk... %ld bytes", dsize );

      if(dsize) memcpy( ctrl->spfileptr + ctrl->bytes_spooled, data, dsize );

      ctrl->bytes_spooled += dsize;
      if( ctrl->bytes_spooled >= ctrl->total_size ) {

	ctrl->flags  &= ~GWP_REMOTE;  /* Clear REMOTE flag       */
	ctrl->cflags |= GWC_DONE;     /* Signify that we're done */

	if( ctrl->cflags & GWC_PORTAL ) {
	  gw_logf( LOG_INTDBG, "Sending spoolfile to scheduler" );

	  if( gwp_send_to_portal( ctrl, ctrl->spoolfile, 
				  strlen(ctrl->spoolfile) + 1 ) < 0 ) {
	    gw_logf( LOG_SYSERR,"Send_to_portal failed with code %d",gwerror);
	    return Failure;
	  }
	} else {
	  gw_logf( LOG_INTDBG, "Sending spoolfile to service/client" );

	  /* This may return < 0 and gwerror = GWE_SMSGWBLK, in which case we
	   * need to try the send again later
	   */
	  if( gwp_send_message( ctrl, ctrl->spoolfile, 
				strlen(ctrl->spoolfile) + 1 ) < 0 )
	    return Failure;
	}
      }
    } else {
      /* Not spooling data, and not going via gwportal... 
       * Therefore the data *must* fit within one msg block!
       */
      if( dsize > ctrl->bsize ) {
	gw_logf( LOG_SERIOUS,"Too many bytes to send in message (%ld)",dsize);
	gwerror = GWE_SYSFAIL;
	return Failure;
      }
      
      gw_logf( LOG_INTDBG, "1: Sending Datablock.... (msg)" );
      
      /* This may return < 0 and gwerror = GWE_SMSGWBLK, in which case we
       * need to try the send again later
       */
      if( gwp_send_message( ctrl, data, dsize ) < 0 ) return Failure;
      ctrl->cflags |= GWC_DONE;     /* Signify that we're done */
    }
  }

  return Success;
}

/* ------------------------------------------------------------------------- */

RetValue
gwp_write_block_flush( GWMsgCtrl *ctrl )
{
  RetValue ret    = Success;
  long     obsize = 0;

  /* force a flush of the cache etc... */
  obsize = ctrl->bsize;
  if( (ctrl->bsize = CSLen( ctrl->cache )) > 0 ) {
    ret = gwp_write_block_data( ctrl, "", 0 );
  }
  ctrl->bsize = obsize;

  /* Now clean up cache, unmap spool file etc...
   */
  free_cstring( &(ctrl->cache) );

  if( ctrl->cflags & GWC_SPOOL && ctrl->spfileptr != NULL ) {
    munmap( ctrl->spfileptr, ctrl->total_size );
    ctrl->spfileptr = NULL;

    if(ret == Failure) unlink(ctrl->spoolfile); /* Unlink if we've an error */
  }

  ctrl->bytes_spooled = 0;
  
  return ret;
}

/* ------------------------------------------------------------------------- */
/* Do NOT use GW_NOBLOCK!
 *
 * Replies will back-up in the queue!
 *
 * SIDE EFFECTS:
 *
 *  ctrl.msgid = ID of Reply Packet received
 *  ctrl.flags = <undefined>
 */
RetValue
gwp_get_reply( GWMsgCtrl *ctrl )
{
  int          fd    = 0;
  CString      tmpCS = CSInit;
  MsgBlock     mb    = {0};
  long         msgtype = 0;
  uuid_t       want_mid;
  int          msg_match = 0;
  mode_t       om;

  if( ctrl->flags & GW_NOREPLY ) {
    gw_logf( LOG_SYSWARN, 
	     "GW_NOREPLY was specified for this service call. "
	     "Cannot get reply." );
    return Failure; /* No Reply! */
  }

  if( ctrl->cflags & GWC_PORTAL ) {
    /* Read Reply straight from gwportal */

    if( gwp_get_reply_from_portal( ctrl ) == NULL ) return Failure;

  } else {
    do {
      errno   = 0;
      msgtype = (ctrl->ident < 0) ? ctrl->ident : -(ctrl->ident); /* Get -ve */
      
      if( msgrcv( ctrl->r_msgqid, (struct msgbuf*)&mb, 
		  sizeof(MsgBlock) - sizeof(long), msgtype, 
		  (ctrl->flags & GW_NOBLOCK) ? IPC_NOWAIT : 0 ) < 0) {
	switch( errno ) {
	case EINTR:
	  break;
	  
	case EIDRM: /* Queue has been removed */
	  gw_logf( LOG_SERIOUS,
		   "Reply receive. Queue %ld removed", ctrl->r_msgqid );
	  free_cstring( &ctrl->cache );   
	  
	  return Failure;
	  break;

	case ENOMSG: /* No message available */
	  /* The caller must check errno for ENOMSG to see that this wasn't
	   * a fatal error....
	   */
	  return Failure;
	  break;
	  
	default:
	  gw_logf( LOG_SERIOUS, "Error receiving reply (error: %d)", errno );
	  free_cstring( &ctrl->cache );      
	  return Failure;
	  break;
	}
      }
    } while( errno == EINTR ); /* Don't let this syscall be interupted! */

    uuid_copy( want_mid,    ctrl->msgid );
    uuid_copy( ctrl->msgid, mb.msgid    );
    ctrl->flags = mb.flags;

    if( gw_log_at_level(LOG_INTDBG) ) {
      char uuidbuf[64];
      gw_logf( LOG_INTDBG, "Check MSGID:" );
      uuid_unparse( want_mid, uuidbuf );
      gw_logf( LOG_INTDBG, "      Want : %s", uuidbuf );
      uuid_unparse( mb.msgid, uuidbuf );
      gw_logf( LOG_INTDBG, "      Got  : %s", uuidbuf );

    }

    msg_match = gwuuid_compare( want_mid, mb.msgid );
	
    if( msg_match != 0 ) gw_logf( LOG_INTDBG, "Collected Stale Reply" );

    if( mb.flags & GWP_DFILE ) { /* Reply is in a spool file */
      
      /* Have we just received the reply we were expecting ??? */
      if( msg_match == 0 ) {
	
	ctrl->cflags   = GWC_SPOOL;
	ctrl->flags   &= ~GWP_DFILE;
	
	om = umask( _Mask );
	if( (fd = open( mb.data, O_RDONLY, 0666 )) < 0 ) {
	  umask(om);
	  gw_logf( LOG_SYSERR, "Cannot open spoolfile '%s' (errno: %ld)",
		   mb.data, (long)errno );
	  return Failure;
	}
	umask(om);
	strcpy( ctrl->spoolfile, mb.data );
	ctrl->total_size = mb.total_size;

	/* Memmap the file
	 */
	gw_logf( LOG_INTDBG, 
		 "Mapping spoolfile '%s' Len %ld", mb.data, ctrl->total_size );
	ctrl->spfileptr = (unsigned char *)mmap( 0, ctrl->total_size,
						 PROT_READ|PROT_WRITE, 
						 MAP_PRIVATE, fd, 0 );
	close(fd);
	
	if( ctrl->spfileptr < 0 ) {
	  unlink( ctrl->spoolfile );
	  ctrl->spfileptr = NULL;
	  gw_logf( LOG_SERIOUS, "mmap failure (errno %ld)", (long)errno );
	  return Failure;
	}
      
	free_cstring( &ctrl->cache );
	WrapCStringLen( ctrl->cache, ctrl->spfileptr, ctrl->total_size );
      } else {
	/* Not our reply - UNLINK the file */
	unlink( mb.data );
      }

      /* Return pointer to mapped data.
       * The caller MUST unmap this file!
       */
      return Success;
    }

    reset_cstring( &ctrl->cache );

    WrapCStringLen( tmpCS, mb.data, mb.dsize );
    append_cstring( &ctrl->cache, &tmpCS );
  }

  return Success;
}

/* ------------------------------------------------------------------------- */
/* This may be called repeatedly on the same GWMsgCtrl without ill effect
 */
RetValue
gwp_cleanup_ctrl( GWMsgCtrl *ctrl )
{
  /* Unmap file if mapped
   */
  if( ctrl->cflags & GWC_SPOOL ) {

    ctrl->flags  &= ~GWP_DFILE;
    ctrl->cflags &= ~GWC_SPOOL;

    if( ctrl->spfileptr != NULL ) {
      munmap( ctrl->spfileptr, ctrl->total_size );
      ctrl->spfileptr = NULL;
    }

    if( (!(ctrl->flags & GWP_REPLY)) || (ctrl->cflags & GWC_FORCERM) ) {
      /* If we're not a reply or we're being told to force a removal
       * then we can unlink the spool file
       */
      unlink( ctrl->spoolfile );  /* Only unlink spool file if not a reply */
    }
  }

  /* Free ctrl DATA if malloced
   */
  free_cstring( &ctrl->cache );

  return Success;
}

/* ------------------------------------------------------------------------- */

void
gwp_remove_reply_queue()
{
  if( _ReplyQueue >= 0 ) {
    gw_logf( LOG_INTDBG, "Remove reply queue %ld", _ReplyQueue );

    msgctl( _ReplyQueue, IPC_RMID, NULL );
    _ReplyQueue = -1;
    if( _MyServEntry ) _MyServEntry->bb_sc.repmqid = -1; /* Servers ONLY */
  }
}

/* ------------------------------------------------------------------------- */

RetValue
gwp_finish_request( GWMsgCtrl *ctrl )
{
  /* unmap files and fre cache memory
   */
  gwp_cleanup_ctrl( ctrl );
  
  /* TODO: Check that we've no outstanding ASYNC replies.... */
  return Success;
}


/* ------------------------------------------------------------------------- */
/*
 * Service's call this to receive their request data
 */
unsigned char *
gwp_get_request( GWMsgCtrl *ctrl, long mtype  )
{
  int             fd     = 0;
  CString         tmpCS  = CSInit;
  int             noblock= 0;
  int             b      = 0;
  int             state  = 0;
  int        timed_block = 0;
  static MsgBlock mb     = {0};

  noblock = ctrl->cflags & GWC_NOBLOCK;
  
  if( ctrl->cflags & GWC_TIMEBLK ) {
    /* We're being asked to block for a limited length of time
     */
    noblock     = 0; /* Disable implicit no-blocking mode */
    timed_block = 1; /* Enable Timed Block                */
  }
  
  ctrl->cflags = 0; /* Clear Control Flags */

  state = TIMER_START;
  
  do {    
    errno = 0;
    if( timed_block ) {
      signal( SIGALRM, __got_timeout );
      state = sigsetjmp( timeout_env, 1 );
    }
    
    if( state == TIMER_EXPIRED ) {
      ctrl->cflags |= GWC_NOREQM; /* Indicate that no request is returned */
      signal( SIGALRM, SIG_IGN );
      return NULL;
    }
    
    if( timed_block ) alarm( ReqBlockTime );
    b = msgrcv( ctrl->msgqid, (struct msgbuf*)&mb,
		sizeof(MsgBlock) - sizeof(long), mtype, noblock?IPC_NOWAIT:0);
    
    if( timed_block ) { 
      signal( SIGALRM, SIG_IGN );
      alarm(0);      
    }
    
    if( b < 0 ) {
      switch( errno ) {
      case EINTR: /* Skip this error as we pick it up in the while loop */
	break;
	
      case EIDRM: /* Queue has been removed */
	gw_logf( LOG_SERIOUS, "Queue %ld has been removed.", ctrl->msgqid );
	/* ARG: Set DO NOT RESPAWN flag... */
	exit(1);
	break;
	
      case ENOMSG: /* IPC_NOWAIT returned no message -> send status back */
	ctrl->cflags |= GWC_NOREQM;
	return NULL;
	break;
	
      default:
	gw_logf( LOG_SERIOUS, "Error receiving request (error: %d)", errno );
	return NULL;
	break;
      }
    }
  } while( errno == EINTR ); /* Don't let this syscall be interupted! */
  
  gw_log_init( 0, mb.service ); /* re-init with service name */
  
  gw_logf( LOG_INTDBG, "Received a request for svc '%s' from client %ld",
	   mb.service, mb.clid );
  
  ctrl->ident        = mb.ident;  
  ctrl->flags        = mb.flags;
  uuid_copy( ctrl->msgid, mb.msgid );
  ctrl->clid         = mb.clid;
  ctrl->service      = mb.service;  /* mb MUST BE STATIC! */
  ctrl->function     = mb.function; /* mb MUST BE STATIC! */
  ctrl->total_size   = mb.total_size;
  ctrl->r_msgqid     = mb.rep_msgqid;
  ctrl->org_r_msgqid = mb.org_rep_msgqid;
  ctrl->destsrvid    = mb.dest_srv_id;
  ctrl->target_dom   = mb.dest_dom_id;
  ctrl->msgtype      = mb.msgtype;


  if( mb.flags & GWP_DFILE ) { /* Request is in a spool file */
    ctrl->cflags   = GWC_SPOOL;
    ctrl->flags   &= ~GWP_DFILE;

    if( (fd = open( mb.data, O_RDONLY, 0666 )) < 0 ) {
      gw_logf( LOG_SYSERR, "Cannot open spoolfile '%s' (errno: %ld)",
	       mb.data, (long)errno );
      return NULL;
    }
    strcpy( ctrl->spoolfile, mb.data );

    /* Memmap the file
     */
    gw_logf( LOG_INTDBG, 
	     "Mapping spoolfile '%s' Len %ld", mb.data, ctrl->total_size );
    ctrl->spfileptr = (unsigned char *)mmap( 0, ctrl->total_size,
					     PROT_READ|PROT_WRITE, 
					     MAP_PRIVATE, fd, 0 );
    close(fd);
    
    if( ctrl->spfileptr < 0 ) {
      unlink( ctrl->spoolfile );
      ctrl->spfileptr = NULL;
      gw_logf( LOG_SERIOUS, "mmap failure (errno %ld)", (long)errno );
      return NULL;
    }
    
    free_cstring( &ctrl->cache );
    WrapCStringLen( ctrl->cache, ctrl->spfileptr, ctrl->total_size );

    /* Return pointer to mapped data.
     * The caller MUST unmap this file!
     */
    return ctrl->spfileptr;
  }
  
  /* Data is carried within a single message block, so we can just reset and
   * use the data cache we have...
   */
  reset_cstring( &ctrl->cache );

  /* Make sure we have a pointer to some data, even if it's of length zero */
  if( CSStart( ctrl->cache ) == NULL ) WrapCStringLen( ctrl->cache, "", 0 );
  
  WrapCStringLen( tmpCS, mb.data, mb.dsize );
  append_cstring( &ctrl->cache, &tmpCS );
  
  /*
   * Pass back pointers to read data....
   * It is the callers responsibility to free these!
   */
  return CSStart( ctrl->cache );
}

/* ------------------------------------------------------------------------- */
/*
 * Does any cleanup required for shuttion down server.
 * Expects the caller to do an immediate exit.
 */
void
gwp_shutdown_server()
{
  /* Wait for between 200 and 800ms depending on pid.
   * This tries to avoid flooding guardian with child death signals!
   */
  usleep( ( (getpid() % 4) + 1 ) * 200 );
  
  /* Remove our reply queue if we've got one remaining...
   */
  if( _ReplyQueue >= 0 ) gwp_remove_reply_queue();

  gw_logf( LOG_AUDIT, "Server %ld exiting...", _ServId );
}

/* ------------------------------------------------------------------------- */
#define GWUCMP(l1,l2) if( *(l1) != *(l2) ) return ( (*(l1) < *(l2)) ? -1 : 1 );
#define DBUG(l1,l2) gw_logf( LOG_INTDBG, "compare %x(%x) with %x(%x)", (l1), *(l1), (l2), *(l2) );

int
gwuuid_compare( uuid_t u1, uuid_t u2 )
{
  int *l1 = (int*)u1;
  int *l2 = (int*)u2;
  /*DBUG( l1+0, l2+0 );*/
  /*DBUG( l1+1, l2+1 );*/
  /*DBUG( l1+2, l2+2 );*/
  /*DBUG( l1+3, l2+3 );*/
  GWUCMP( l1+0, l2+0 );
  GWUCMP( l1+1, l2+1 );
  GWUCMP( l1+2, l2+2 );
  GWUCMP( l1+3, l2+3 );

  return 0;
}

/* ------------------------------------------------------------------------- */
/* end */
