/*
 * Goldwater Distributed Web Architecture.
 *
 * XML parsing tokens
 *
 * Copyright 1996-2002 netFluid Technology Ltd
 *
 * Goldwater is written and maintained by netFluid Technology Ltd.
 * http://www.nfluid.co.uk
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 *
 * $Revision: 1.2 $  $Date: 2003/03/29 12:21:11 $
 * 
 * ---------------------------------------------------------------------------
 */
#ifndef GW_COMPILE_LOG_H
#define GW_COMPILE_LOG_H

#include <phlib/xl_linklst.h>
#include <phlib/xl_string.h>

#define URI_MAX       256
#define SERVICE_MAX   256
#define MAX_TAG_DEPTH 255 /* Max number of nested tags... */

static  const int TAG_UNKNOWN = 32768;

/* ------------------------------------------------------------------------- */
/* Internal Flags
 */
typedef enum {
  F_TAG_OPEN     = 0x01,
  F_TAG_CLOSE    = 0x02,
  F_TAG_EMBEDDED = 0x04
} InternalFlags;


typedef struct {
  int     token;
  CString elementCS;
  CString dataCS;
  long    balanced;
  long    data_SP; /* Stack node to write 'freetext' data to */
} XMLTag;

typedef struct {
  long chtag;
  char *xmltag;
} XMLTagMap;


typedef struct {
  char *t_text;
  int   token;
} TokenMap;

/* ------------------------------------------------------------------------- */
/*
 * Define XML parsing structs
 */
typedef enum {
  T_TERMINATE   = 0x00,
  T_GOLDWATER   = 0x01,
  T_CONTROL     = 0x02,
  T_LOCAL       = 0x03,
  T_REMOTE      = 0x04,
  T_DOMAIN      = 0x05,
  T_NAME        = 0x06,
  T_NS          = 0x07,
  T_ADDR        = 0x08,
  T_PORT        = 0x09,
  T_EXEDIR      = 0x0a,
  T_LOGFILE     = 0x0b,
  T_LOGLEVEL    = 0x0c,
  T_MODE        = 0x0d,
  T_IPCKEY      = 0x0e,
  T_REQTOUT     = 0x0f,
  T_MAXSERVERS  = 0x10,
  T_MAXSERVICES = 0x11,
  T_DEFAULT     = 0x12,
  T_MIN         = 0x13,
  T_MAX         = 0x14,
  T_PRIORITY    = 0x15,
  T_SERVERS     = 0x16,
  T_SERVICE     = 0x17,
  T_FUNCTION    = 0x18,
  T_CLASS       = 0x19,
  T_ID          = 0x1a,
  T_EXE         = 0x1b,
  T_MAXDOMAINS  = 0x1c,
  T_SPOOLDIR    = 0x1d,
  T_STDOUT      = 0x1e,
  T_STDERR      = 0x1f,
  T_IMPORT      = 0x20,
  T_EXPORT      = 0x21,
  T_MAXCLIENTS  = 0x22,
  T_ENVIRONMENT = 0x23,
  T_USER        = 0x24,
  T_GROUP       = 0x25,
  T_UMASK       = 0x26,
  T_SANDBOX     = 0x27,
  T_MAXREQUESTS = 0x28,

  MAX_TOKEN     = 0x28
} TagToken;

/* ------------------------------------------------------------------------- */

TokenMap Mapping[] = {
  { "goldwater",   T_GOLDWATER },
  { "control",     T_CONTROL },
  { "local",       T_LOCAL },
  { "remote",      T_REMOTE },
  { "domain",      T_DOMAIN },
  { "name",        T_NAME },
  { "ns",          T_NS },
  { "addr",        T_ADDR },
  { "port",        T_PORT },
  { "user",        T_USER },
  { "group",       T_GROUP },
  { "exedir",      T_EXEDIR },
  { "logfile",     T_LOGFILE },
  { "loglevel",    T_LOGLEVEL },
  { "mode",        T_MODE },
  { "umask",       T_UMASK },
  { "ipckey",      T_IPCKEY },
  { "reqtout",     T_REQTOUT },
  { "maxservers",  T_MAXSERVERS },
  { "maxclients",  T_MAXCLIENTS },
  { "maxservices", T_MAXSERVICES },
  { "default",     T_DEFAULT },
  { "min",         T_MIN },
  { "max",         T_MAX },
  { "maxrequests", T_MAXREQUESTS },
  { "priority",    T_PRIORITY },
  { "servers",     T_SERVERS },
  { "env",         T_ENVIRONMENT },
  { "service",     T_SERVICE },
  { "import",      T_IMPORT },
  { "export",      T_EXPORT },
  { "function",    T_FUNCTION },
  { "class",       T_CLASS },
  { "id",          T_ID },
  { "exe",         T_EXE },
  { "maxdomains",  T_MAXDOMAINS },
  { "spooldir",    T_SPOOLDIR },
  { "stdout",      T_STDOUT },
  { "stderr",      T_STDERR },
  { "sandbox",     T_SANDBOX },
  { NULL, T_TERMINATE }
};


/* ------------------------------------------------------------------------- */
/* Prototypes */

/* ------------------------------------------------------------------------- */
#endif

